# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>, Lars Norpchen"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import datetime
import logging
import subprocess

try:
    import queue
except ImportError:
    import Queue as queue

import collections
import re
import threading

import octoprint.plugin
from octoprint.settings import settings

# singleton
_instance = None


def all_events():
    return [
        getattr(Events, name)
        for name in Events.__dict__
        if not name.startswith("_") and name not in ("register_event",)
    ]


class Events(object):
    # server
    STARTUP = "Startup"
    SHUTDOWN = "Shutdown"
    CONNECTIVITY_CHANGED = "ConnectivityChanged"

    # connect/disconnect to printer
    CONNECTING = "Connecting"
    CONNECTED = "Connected"
    DISCONNECTING = "Disconnecting"
    DISCONNECTED = "Disconnected"

    # State changes
    PRINTER_STATE_CHANGED = "PrinterStateChanged"
    PRINTER_RESET = "PrinterReset"

    # connect/disconnect by client
    CLIENT_OPENED = "ClientOpened"
    CLIENT_CLOSED = "ClientClosed"
    CLIENT_AUTHED = "ClientAuthed"
    CLIENT_DEAUTHED = "ClientDeauthed"

    # user login/logout
    USER_LOGGED_IN = "UserLoggedIn"
    USER_LOGGED_OUT = "UserLoggedOut"

    # File management
    UPLOAD = "Upload"
    FILE_SELECTED = "FileSelected"
    FILE_DESELECTED = "FileDeselected"
    UPDATED_FILES = "UpdatedFiles"
    METADATA_ANALYSIS_STARTED = "MetadataAnalysisStarted"
    METADATA_ANALYSIS_FINISHED = "MetadataAnalysisFinished"
    METADATA_STATISTICS_UPDATED = "MetadataStatisticsUpdated"

    FILE_ADDED = "FileAdded"
    FILE_REMOVED = "FileRemoved"
    FOLDER_ADDED = "FolderAdded"
    FOLDER_REMOVED = "FolderRemoved"

    # SD Upload
    TRANSFER_STARTED = "TransferStarted"
    TRANSFER_DONE = "TransferDone"
    TRANSFER_FAILED = "TransferFailed"

    # print job
    PRINT_STARTED = "PrintStarted"
    PRINT_DONE = "PrintDone"
    PRINT_FAILED = "PrintFailed"
    PRINT_CANCELLING = "PrintCancelling"
    PRINT_CANCELLED = "PrintCancelled"
    PRINT_PAUSED = "PrintPaused"
    PRINT_RESUMED = "PrintResumed"
    ERROR = "Error"

    # print/gcode events
    POWER_ON = "PowerOn"
    POWER_OFF = "PowerOff"
    HOME = "Home"
    Z_CHANGE = "ZChange"
    WAITING = "Waiting"
    DWELL = "Dwelling"
    COOLING = "Cooling"
    ALERT = "Alert"
    CONVEYOR = "Conveyor"
    EJECT = "Eject"
    E_STOP = "EStop"
    POSITION_UPDATE = "PositionUpdate"
    FIRMWARE_DATA = "FirmwareData"
    TOOL_CHANGE = "ToolChange"
    REGISTERED_MESSAGE_RECEIVED = "RegisteredMessageReceived"
    COMMAND_SUPPRESSED = "CommandSuppressed"
    INVALID_TOOL_REPORTED = "InvalidToolReported"

    # Timelapse
    CAPTURE_START = "CaptureStart"
    CAPTURE_DONE = "CaptureDone"
    CAPTURE_FAILED = "CaptureFailed"
    POSTROLL_START = "PostRollStart"
    POSTROLL_END = "PostRollEnd"
    MOVIE_RENDERING = "MovieRendering"
    MOVIE_DONE = "MovieDone"
    MOVIE_FAILED = "MovieFailed"

    # Slicing
    SLICING_STARTED = "SlicingStarted"
    SLICING_DONE = "SlicingDone"
    SLICING_FAILED = "SlicingFailed"
    SLICING_CANCELLED = "SlicingCancelled"
    SLICING_PROFILE_ADDED = "SlicingProfileAdded"
    SLICING_PROFILE_MODIFIED = "SlicingProfileModified"
    SLICING_PROFILE_DELETED = "SlicingProfileDeleted"

    # Printer Profiles
    PRINTER_PROFILE_ADDED = "PrinterProfileAdded"
    PRINTER_PROFILE_MODIFIED = "PrinterProfileModified"
    PRINTER_PROFILE_DELETED = "PrinterProfileDeleted"

    # Settings
    SETTINGS_UPDATED = "SettingsUpdated"

    @classmethod
    def register_event(cls, event, prefix=None):
        name = cls._to_identifier(event)
        if prefix:
            event = prefix + event
            name = cls._to_identifier(prefix) + name
        setattr(cls, name, event)
        return name, event

    # based on https://stackoverflow.com/a/1176023
    _first_cap_re = re.compile("(.)([A-Z][a-z]+)")
    _all_cap_re = re.compile("([a-z0-9])([A-Z])")

    @classmethod
    def _to_identifier(cls, name):
        s1 = cls._first_cap_re.sub(r"\1_\2", name)
        return cls._all_cap_re.sub(r"\1_\2", s1).upper()


def eventManager():
    global _instance
    if _instance is None:
        _instance = EventManager()
    return _instance


class EventManager(object):
    """
    Handles receiving events and dispatching them to subscribers
    """

    def __init__(self):
        self._registeredListeners = collections.defaultdict(list)
        self._logger = logging.getLogger(__name__)
        self._logger_fire = logging.getLogger("{}.fire".format(__name__))

        self._startup_signaled = False
        self._shutdown_signaled = False

        self._queue = queue.Queue()
        self._held_back = queue.Queue()

        self._worker = threading.Thread(target=self._work)
        self._worker.daemon = True
        self._worker.start()

    def _work(self):
        try:
            while not self._shutdown_signaled:
                event, payload = self._queue.get(True)
                if event == Events.SHUTDOWN:
                    # we've got the shutdown event here, stop event loop processing after this has been processed
                    self._logger.info(
                        "Processing shutdown event, this will be our last event"
                    )
                    self._shutdown_signaled = True

                eventListeners = self._registeredListeners[event]
                self._logger_fire.debug(
                    "Firing event: {} (Payload: {!r})".format(event, payload)
                )

                for listener in eventListeners:
                    self._logger.debug("Sending action to {!r}".format(listener))
                    try:
                        listener(event, payload)
                    except Exception:
                        self._logger.exception(
                            "Got an exception while sending event {} (Payload: {!r}) to {}".format(
                                event, payload, listener
                            )
                        )

                octoprint.plugin.call_plugin(
                    octoprint.plugin.types.EventHandlerPlugin,
                    "on_event",
                    args=(event, payload),
                )
            self._logger.info("Event loop shut down")
        except Exception:
            self._logger.exception("Ooops, the event bus worker loop crashed")

    def fire(self, event, payload=None):
        """
        Fire an event to anyone subscribed to it

        Any object can generate an event and any object can subscribe to the event's name as a string (arbitrary, but
        case sensitive) and any extra payload data that may pertain to the event.

        Callbacks must implement the signature "callback(event, payload)", with "event" being the event's name and
        payload being a payload object specific to the event.
        """

        send_held_back = False
        if event == Events.STARTUP:
            self._logger.info("Processing startup event, this is our first event")
            self._startup_signaled = True
            send_held_back = True

        self._enqueue(event, payload)

        if send_held_back:
            self._logger.info(
                "Adding {} events to queue that "
                "were held back before startup event".format(self._held_back.qsize())
            )
            while True:
                try:
                    self._queue.put(self._held_back.get(block=False))
                except queue.Empty:
                    break

    def _enqueue(self, event, payload):
        if self._startup_signaled:
            q = self._queue
        else:
            q = self._held_back

        q.put((event, payload))

    def subscribe(self, event, callback):
        """
        Subscribe a listener to an event -- pass in the event name (as a string) and the callback object
        """

        if callback in self._registeredListeners[event]:
            # callback is already subscribed to the event
            return

        self._registeredListeners[event].append(callback)
        self._logger.debug(
            "Subscribed listener {!r} for event {}".format(callback, event)
        )

    def unsubscribe(self, event, callback):
        """
        Unsubscribe a listener from an event -- pass in the event name (as string) and the callback object
        """

        try:
            self._registeredListeners[event].remove(callback)
        except ValueError:
            # not registered
            pass

    def join(self, timeout=None):
        self._worker.join(timeout)
        return self._worker.is_alive()


class GenericEventListener(object):
    """
    The GenericEventListener can be subclassed to easily create custom event listeners.
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

    def subscribe(self, events):
        """
        Subscribes the eventCallback method for all events in the given list.
        """

        for event in events:
            eventManager().subscribe(event, self.eventCallback)

    def unsubscribe(self, events):
        """
        Unsubscribes the eventCallback method for all events in the given list
        """

        for event in events:
            eventManager().unsubscribe(event, self.eventCallback)

    def eventCallback(self, event, payload):
        """
        Actual event callback called with name of event and optional payload. Not implemented here, override in
        child classes.
        """
        pass


class DebugEventListener(GenericEventListener):
    def __init__(self):
        GenericEventListener.__init__(self)

        events = list(filter(lambda x: not x.startswith("__"), dir(Events)))
        self.subscribe(events)

    def eventCallback(self, event, payload):
        GenericEventListener.eventCallback(self, event, payload)
        self._logger.debug("Received event: %s (Payload: %r)" % (event, payload))


class CommandTrigger(GenericEventListener):
    def __init__(self, printer):
        GenericEventListener.__init__(self)
        self._printer = printer
        self._subscriptions = {}

        self._initSubscriptions()

    def _initSubscriptions(self):
        """
        Subscribes all events as defined in "events > $triggerType > subscriptions" in the settings with their
        respective commands.
        """
        if not settings().get(["events"]):
            return

        if not settings().getBoolean(["events", "enabled"]):
            return

        eventsToSubscribe = []
        subscriptions = settings().get(["events", "subscriptions"])
        for subscription in subscriptions:
            if not isinstance(subscription, dict):
                self._logger.info(
                    "Invalid subscription definition, not a dictionary: {!r}".format(
                        subscription
                    )
                )
                continue

            if (
                "event" not in subscription
                or "command" not in subscription
                or "type" not in subscription
                or subscription["type"] not in ["system", "gcode"]
            ):
                self._logger.info(
                    "Invalid command trigger, missing either event, type or command or type is invalid: {!r}".format(
                        subscription
                    )
                )
                continue

            if "enabled" in subscription and not subscription["enabled"]:
                self._logger.info("Disabled command trigger: {!r}".format(subscription))
                continue

            events = subscription["event"]
            command = subscription["command"]
            commandType = subscription["type"]
            debug = subscription["debug"] if "debug" in subscription else False

            # "event" in the configuration can be a string, or
            # a list of strings.  If it's the former, convert it
            # into the latter.
            if not isinstance(events, (tuple, list, set)):
                events = [events]

            for event in events:
                if event not in self._subscriptions:
                    self._subscriptions[event] = []
                self._subscriptions[event].append((command, commandType, debug))

                if event not in eventsToSubscribe:
                    eventsToSubscribe.append(event)

        self.subscribe(eventsToSubscribe)

    def eventCallback(self, event, payload):
        """
        Event callback, iterates over all subscribed commands for the given event, processes the command
        string and then executes the command via the abstract executeCommand method.
        """

        GenericEventListener.eventCallback(self, event, payload)

        if event not in self._subscriptions:
            return

        for command, commandType, debug in self._subscriptions[event]:
            try:
                if isinstance(command, (tuple, list, set)):
                    processedCommand = []
                    for c in command:
                        processedCommand.append(self._processCommand(c, event, payload))
                else:
                    processedCommand = self._processCommand(command, event, payload)
                self.executeCommand(processedCommand, commandType, debug=debug)
            except KeyError:
                self._logger.warning(
                    "There was an error processing one or more placeholders in the following command: %s"
                    % command
                )

    def executeCommand(self, command, commandType, debug=False):
        if commandType == "system":
            self._executeSystemCommand(command, debug=debug)
        elif commandType == "gcode":
            self._executeGcodeCommand(command, debug=debug)

    def _executeSystemCommand(self, command, debug=False):
        def commandExecutioner(cmd):
            if debug:
                self._logger.info("Executing system command: {}".format(cmd))
            else:
                self._logger.info("Executing a system command")
            # we run this with shell=True since we have to trust whatever
            # our admin configured as command and since we want to allow
            # shell-alike handling here...
            subprocess.check_call(cmd, shell=True)

        def process():
            try:
                if isinstance(command, (list, tuple, set)):
                    for c in command:
                        commandExecutioner(c)
                else:
                    commandExecutioner(command)
            except subprocess.CalledProcessError as e:
                if debug:
                    self._logger.warning(
                        "Command failed with return code {}: {}".format(
                            e.returncode, str(e)
                        )
                    )
                else:
                    self._logger.warning(
                        "Command failed with return code {}, enable debug logging on target 'octoprint.events' for details".format(
                            e.returncode
                        )
                    )
            except Exception:
                self._logger.exception("Command failed")

        t = threading.Thread(target=process)
        t.daemon = True
        t.start()

    def _executeGcodeCommand(self, command, debug=False):
        commands = [command]
        if isinstance(command, (list, tuple, set)):
            commands = list(command)
        if debug:
            self._logger.info("Executing GCode commands: %r" % command)
        self._printer.commands(commands)

    def _processCommand(self, command, event, payload):
        """
        Performs string substitutions in the command string based on a few current parameters.

        The following substitutions are currently supported:

          - {__currentZ} : current Z position of the print head, or -1 if not available
          - {__eventname} : the name of the event hook being triggered
          - {__filename} : name of currently selected file, or "NO FILE" if no file is selected
          - {__filepath} : path in origin location of currently selected file, or "NO FILE" if no file is selected
          - {__fileorigin} : origin of currently selected file, or "NO FILE" if no file is selected
          - {__progress} : current print progress in percent, 0 if no print is in progress
          - {__data} : the string representation of the event's payload
          - {__json} : the json representation of the event's payload, "{}" if there is no payload, "" if there was an error on serialization
          - {__now} : ISO 8601 representation of the current date and time

        Additionally, the keys of the event's payload can also be used as placeholder.
        """

        json_string = "{}"
        if payload:
            import json

            try:
                json_string = json.dumps(payload)
            except Exception:
                self._logger.exception("JSON: Cannot dump %r", payload)
                json_string = ""

        params = {
            "__currentZ": "-1",
            "__eventname": event,
            "__filename": "NO FILE",
            "__filepath": "NO PATH",
            "__progress": "0",
            "__data": str(payload),
            "__json": json_string,
            "__now": datetime.datetime.now().isoformat(),
        }

        currentData = self._printer.get_current_data()

        if "currentZ" in currentData and currentData["currentZ"] is not None:
            params["__currentZ"] = str(currentData["currentZ"])

        if (
            "job" in currentData
            and "file" in currentData["job"]
            and "name" in currentData["job"]["file"]
            and currentData["job"]["file"]["name"] is not None
        ):
            params["__filename"] = currentData["job"]["file"]["name"]
            params["__filepath"] = currentData["job"]["file"]["path"]
            params["__fileorigin"] = currentData["job"]["file"]["origin"]
            if (
                "progress" in currentData
                and currentData["progress"] is not None
                and "completion" in currentData["progress"]
                and currentData["progress"]["completion"] is not None
            ):
                params["__progress"] = str(round(currentData["progress"]["completion"]))

        # now add the payload keys as well
        if isinstance(payload, dict):
            params.update(payload)

        return command.format(**params)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


def normalize(path, expand_user=True, expand_vars=True, real=True, **kwargs):
    import os

    if path is None:
        return None

    if expand_user:
        path = os.path.expanduser(path)
    if expand_vars:
        path = os.path.expandvars(path)
    path = os.path.abspath(path)
    if real:
        path = os.path.realpath(path)
    return path



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <gina@octoprint.org>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import gzip
import logging
import os
import re

try:
    from urllib import parse as urlparse
except ImportError:
    import urlparse

import webassets.filter.cssrewrite.urlpath as urlpath
from webassets.bundle import Bundle
from webassets.filter import Filter
from webassets.filter.cssrewrite.base import PatternRewriter
from webassets.merge import BaseHunk, MemoryHunk


def replace_url(source_url, output_url, url):
    # If path is an absolute one, keep it
    parsed = urlparse.urlparse(url)

    if not parsed.scheme and not parsed.path.startswith("/"):
        abs_source_url = urlparse.urljoin(source_url, url)

        # relpath() will not detect this case
        if urlparse.urlparse(abs_source_url).scheme:
            return abs_source_url

        # rewritten url: relative path from new location (output)
        # to location of referenced file (source + current url)
        url = urlpath.relpath(output_url, abs_source_url)

    return url


class UrlRewriter(PatternRewriter):
    def input(self, _in, out, **kw):
        source, source_path, output, output_path = (
            kw["source"],
            kw["source_path"],
            kw["output"],
            kw["output_path"],
        )

        self.source_path = source_path
        self.output_path = output_path
        self.source_url = self.ctx.resolver.resolve_source_to_url(
            self.ctx, source_path, source
        )
        self.output_url = self.ctx.resolver.resolve_output_to_url(self.ctx, output)

        return super(UrlRewriter, self).input(_in, out, **kw)

    def replace_url(self, url):
        return replace_url(self.source_url, self.output_url, url)


class LessImportRewrite(UrlRewriter):
    name = "less_importrewrite"

    patterns = {"import_rewrite": re.compile(r"(@import(\s+\(.*\))?\s+)\"(.*)\";")}

    def import_rewrite(self, m):
        import_with_options = m.group(1)
        import_url = self.replace_url(m.group(3))
        return '{import_with_options}"{import_url}";'.format(**locals())


class SourceMapRewrite(UrlRewriter):
    name = "sourcemap_urlrewrite"

    patterns = {"url_rewrite": re.compile(r"(//#\s+sourceMappingURL=)(.*)")}

    def url_rewrite(self, m):
        mapping = m.group(1)
        source_url = self.replace_url(m.group(2))
        return "{mapping}{source_url}".format(**locals())


class SourceMapRemove(PatternRewriter):
    name = "sourcemap_remove"

    patterns = {"sourcemap_remove": re.compile(r"(//#\s+sourceMappingURL=)(.*)")}

    def sourcemap_remove(self, m):
        return ""


class JsDelimiterBundler(Filter):
    name = "js_delimiter_bundler"
    options = {}

    def input(self, _in, out, **kwargs):
        source = kwargs.get("source", "n/a")

        out.write("// source: " + source + "\n")
        out.write(_in.read())
        out.write("\n;\n")


class GzipFile(Filter):
    name = "gzip"
    options = {}

    def output(self, _in, out, **kwargs):
        data = _in.read()
        out.write(data)

        # webassets requires us to output a "str", but we can't do that since gzip
        # provides binary outputs.
        #
        # We work around that by outputting the gzipped file to another path
        output_path = kwargs.get("output_path", None)
        if output_path:
            gzipped_output_path = output_path + ".gz"
            try:
                with gzip.open(gzipped_output_path, "wb", 9) as f:
                    f.write(data.encode("utf8"))
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error writing gzipped "
                    "output of {} to {}".format(output_path, gzipped_output_path)
                )
                try:
                    os.remove(gzipped_output_path)
                except Exception:
                    logging.getLogger(__name__).exception(
                        "Error removing broken " ".gz from {}".format(gzipped_output_path)
                    )


class ChainedHunk(BaseHunk):
    def __init__(self, *hunks):
        self._hunks = hunks

    def mtime(self):
        pass

    def data(self):
        result = ""
        for hunk in self._hunks:
            if isinstance(hunk, tuple) and len(hunk) == 2:
                hunk, f = hunk
            else:
                f = lambda x: x
            result += f(hunk.data())
        return result


_PLUGIN_BUNDLE_WRAPPER_PREFIX = """// JS assets for plugin {plugin}
(function () {{
    try {{
        """
_PLUGIN_BUNDLE_WRAPPER_SUFFIX = """
    }} catch (error) {{
        log.error("Error in JS assets for plugin {plugin}:", (error.stack || error));
    }}
}})();
"""


class JsPluginBundle(Bundle):
    def __init__(self, plugin, *args, **kwargs):
        Bundle.__init__(self, *args, **kwargs)
        self.plugin = plugin

    def _merge_and_apply(
        self,
        ctx,
        output,
        force,
        parent_debug=None,
        parent_filters=None,
        extra_filters=None,
        disable_cache=None,
    ):
        hunk = Bundle._merge_and_apply(
            self,
            ctx,
            output,
            force,
            parent_debug=parent_debug,
            parent_filters=parent_filters,
            extra_filters=extra_filters,
            disable_cache=disable_cache,
        )

        return ChainedHunk(
            MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_PREFIX.format(plugin=self.plugin)),
            (hunk, lambda x: x.replace("\n", "\n        ")),
            MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_SUFFIX.format(plugin=self.plugin)),
        )



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import, division, print_function, unicode_literals

# Version: 0.15+dev

"""
HEADS UP!
=========

You are looking at a heavily modified version of the original
[Versioneer](https://github.com/warner/python-versioneer), customized
to fit the development of [OctoPrint](https://github.com/foosel/OctoPrint).

This version adds

  * support for inclusion of the current branch in the available pieces
    to use for version generation
  * a lookup file to use, to modify version generation based on the
    currently checked out branch including support for virtual tags
    pointing to specific commit hashes
  * a new style pep440-dev: TAG.devDISTANCE[.dirty]+gHEX
  * a new style pep440-tag: TAG[.postDISTANCE.dev0+gHEX]

Please note that this fork is only under maintenance as far as
required by the OctoPrint project. If you are looking for an actively
maintained version of Versioneer, please go to the original project.

The Versioneer
==============

* like a rocketeer, but for versions!
* https://github.com/warner/python-versioneer
* Brian Warner
* License: Public Domain
* Compatible With: python2.6, 2.7, 3.2, 3.3, 3.4, and pypy
* [![Latest Version]
(https://pypip.in/version/versioneer/badge.svg?style=flat)
](https://pypi.python.org/pypi/versioneer/)
* [![Build Status]
(https://travis-ci.org/warner/python-versioneer.png?branch=master)
](https://travis-ci.org/warner/python-versioneer)

This is a tool for managing a recorded version number in distutils-based
python projects. The goal is to remove the tedious and error-prone "update
the embedded version string" step from your release process. Making a new
release should be as easy as recording a new tag in your version-control
system, and maybe making new tarballs.


## Quick Install

* `pip install versioneer` to somewhere to your $PATH
* add a `[versioneer]` section to your setup.cfg (see below)
* run `versioneer install` in your source tree, commit the results

## Version Identifiers

Source trees come from a variety of places:

* a version-control system checkout (mostly used by developers)
* a nightly tarball, produced by build automation
* a snapshot tarball, produced by a web-based VCS browser, like github's
  "tarball from tag" feature
* a release tarball, produced by "setup.py sdist", distributed through PyPI

Within each source tree, the version identifier (either a string or a number,
this tool is format-agnostic) can come from a variety of places:

* ask the VCS tool itself, e.g. "git describe" (for checkouts), which knows
  about recent "tags" and an absolute revision-id
* the name of the directory into which the tarball was unpacked
* an expanded VCS keyword ($Id$, etc)
* a `_version.py` created by some earlier build step

For released software, the version identifier is closely related to a VCS
tag. Some projects use tag names that include more than just the version
string (e.g. "myproject-1.2" instead of just "1.2"), in which case the tool
needs to strip the tag prefix to extract the version identifier. For
unreleased software (between tags), the version identifier should provide
enough information to help developers recreate the same tree, while also
giving them an idea of roughly how old the tree is (after version 1.2, before
version 1.3). Many VCS systems can report a description that captures this,
for example `git describe --tags --dirty --always` reports things like
"0.7-1-g574ab98-dirty" to indicate that the checkout is one revision past the
0.7 tag, has a unique revision id of "574ab98", and is "dirty" (it has
uncommitted changes.

The version identifier is used for multiple purposes:

* to allow the module to self-identify its version: `myproject.__version__`
* to choose a name and prefix for a 'setup.py sdist' tarball

## Theory of Operation

Versioneer works by adding a special `_version.py` file into your source
tree, where your `__init__.py` can import it. This `_version.py` knows how to
dynamically ask the VCS tool for version information at import time.

`_version.py` also contains `$Revision$` markers, and the installation
process marks `_version.py` to have this marker rewritten with a tag name
during the `git archive` command. As a result, generated tarballs will
contain enough information to get the proper version.

To allow `setup.py` to compute a version too, a `versioneer.py` is added to
the top level of your source tree, next to `setup.py` and the `setup.cfg`
that configures it. This overrides several distutils/setuptools commands to
compute the version when invoked, and changes `setup.py build` and `setup.py
sdist` to replace `_version.py` with a small static file that contains just
the generated version data.

## Installation

First, decide on values for the following configuration variables:

* `VCS`: the version control system you use. Currently accepts "git".

* `style`: the style of version string to be produced. See "Styles" below for
  details. Defaults to "pep440", which looks like
  `TAG[+DISTANCE.gSHORTHASH[.dirty]]`.

* `versionfile_source`:

  A project-relative pathname into which the generated version strings should
  be written. This is usually a `_version.py` next to your project's main
  `__init__.py` file, so it can be imported at runtime. If your project uses
  `src/myproject/__init__.py`, this should be `src/myproject/_version.py`.
  This file should be checked in to your VCS as usual: the copy created below
  by `setup.py setup_versioneer` will include code that parses expanded VCS
  keywords in generated tarballs. The 'build' and 'sdist' commands will
  replace it with a copy that has just the calculated version string.

  This must be set even if your project does not have any modules (and will
  therefore never import `_version.py`), since "setup.py sdist" -based trees
  still need somewhere to record the pre-calculated version strings. Anywhere
  in the source tree should do. If there is a `__init__.py` next to your
  `_version.py`, the `setup.py setup_versioneer` command (described below)
  will append some `__version__`-setting assignments, if they aren't already
  present.

* `versionfile_build`:

  Like `versionfile_source`, but relative to the build directory instead of
  the source directory. These will differ when your setup.py uses
  'package_dir='. If you have `package_dir={'myproject': 'src/myproject'}`,
  then you will probably have `versionfile_build='myproject/_version.py'` and
  `versionfile_source='src/myproject/_version.py'`.

  If this is set to None, then `setup.py build` will not attempt to rewrite
  any `_version.py` in the built tree. If your project does not have any
  libraries (e.g. if it only builds a script), then you should use
  `versionfile_build = None` and override `distutils.command.build_scripts`
  to explicitly insert a copy of `versioneer.get_version()` into your
  generated script.

* `tag_prefix`:

  a string, like 'PROJECTNAME-', which appears at the start of all VCS tags.
  If your tags look like 'myproject-1.2.0', then you should use
  tag_prefix='myproject-'. If you use unprefixed tags like '1.2.0', this
  should be an empty string, using either `tag_prefix=` or `tag_prefix=''`.

* `parentdir_prefix`:

  a optional string, frequently the same as tag_prefix, which appears at the
  start of all unpacked tarball filenames. If your tarball unpacks into
  'myproject-1.2.0', this should be 'myproject-'. To disable this feature,
  just omit the field from your `setup.cfg`.

This tool provides one script, named `versioneer`. That script has one mode,
"install", which writes a copy of `versioneer.py` into the current directory
and runs `versioneer.py setup` to finish the installation.

To versioneer-enable your project:

* 1: Modify your `setup.cfg`, adding a section named `[versioneer]` and
  populating it with the configuration values you decided earlier (note that
  the option names are not case-sensitive):

  ````
  [versioneer]
  VCS = git
  style = pep440
  versionfile_source = src/myproject/_version.py
  versionfile_build = myproject/_version.py
  tag_prefix =
  parentdir_prefix = myproject-
  ````

* 2: Run `versioneer install`. This will do the following:

  * copy `versioneer.py` into the top of your source tree
  * create `_version.py` in the right place (`versionfile_source`)
  * modify your `__init__.py` (if one exists next to `_version.py`) to define
    `__version__` (by calling a function from `_version.py`)
  * modify your `MANIFEST.in` to include both `versioneer.py` and the
    generated `_version.py` in sdist tarballs

  `versioneer install` will complain about any problems it finds with your
  `setup.py` or `setup.cfg`. Run it multiple times until you have fixed all
  the problems.

* 3: add a `import versioneer` to your setup.py, and add the following
  arguments to the setup() call:

        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),

* 4: commit these changes to your VCS. To make sure you won't forget,
  `versioneer install` will mark everything it touched for addition using
  `git add`. Don't forget to add `setup.py` and `setup.cfg` too.

## Post-Installation Usage

Once established, all uses of your tree from a VCS checkout should get the
current version string. All generated tarballs should include an embedded
version string (so users who unpack them will not need a VCS tool installed).

If you distribute your project through PyPI, then the release process should
boil down to two steps:

* 1: git tag 1.0
* 2: python setup.py register sdist upload

If you distribute it through github (i.e. users use github to generate
tarballs with `git archive`), the process is:

* 1: git tag 1.0
* 2: git push; git push --tags

Versioneer will report "0+untagged.NUMCOMMITS.gHASH" until your tree has at
least one tag in its history.

## Version-String Flavors

Code which uses Versioneer can learn about its version string at runtime by
importing `_version` from your main `__init__.py` file and running the
`get_versions()` function. From the "outside" (e.g. in `setup.py`), you can
import the top-level `versioneer.py` and run `get_versions()`.

Both functions return a dictionary with different flavors of version
information:

* `['version']`: A condensed version string, rendered using the selected
  style. This is the most commonly used value for the project's version
  string. The default "pep440" style yields strings like `0.11`,
  `0.11+2.g1076c97`, or `0.11+2.g1076c97.dirty`. See the "Styles" section
  below for alternative styles.

* `['full-revisionid']`: detailed revision identifier. For Git, this is the
  full SHA1 commit id, e.g. "1076c978a8d3cfc70f408fe5974aa6c092c949ac".

* `['dirty']`: a boolean, True if the tree has uncommitted changes. Note that
  this is only accurate if run in a VCS checkout, otherwise it is likely to
  be False or None

* `['error']`: if the version string could not be computed, this will be set
  to a string describing the problem, otherwise it will be None. It may be
  useful to throw an exception in setup.py if this is set, to avoid e.g.
  creating tarballs with a version string of "unknown".

If the underlying VCS supports it and that information is available, this will
also be included:

* `['branch']`: A string with the VCS branch name the version was built on.

Some variants are more useful than others. Including `full-revisionid` in a
bug report should allow developers to reconstruct the exact code being tested
(or indicate the presence of local changes that should be shared with the
developers). `version` is suitable for display in an "about" box or a CLI
`--version` output: it can be easily compared against release notes and lists
of bugs fixed in various releases. Augmenting that with the `branch`
information if it is available will give additional hints during bug reporting
what kind of setup a user was running.

The installer adds the following text to your `__init__.py` to place a basic
version in `YOURPROJECT.__version__`:

    from ._version import get_versions
    __version__ = get_versions()['version']
    del get_versions

## Styles

The setup.cfg `style=` configuration controls how the VCS information is
rendered into a version string.

The default style, "pep440", produces a PEP440-compliant string, equal to the
un-prefixed tag name for actual releases, and containing an additional "local
version" section with more detail for in-between builds. For Git, this is
TAG[+DISTANCE.gHEX[.dirty]] , using information from `git describe --tags
--dirty --always`. For example "0.11+2.g1076c97.dirty" indicates that the
tree is like the "1076c97" commit but has uncommitted changes (".dirty"), and
that this commit is two revisions ("+2") beyond the "0.11" tag. For released
software (exactly equal to a known tag), the identifier will only contain the
stripped tag, e.g. "0.11".

Other styles are available. See details.md in the Versioneer source tree for
descriptions.

## Debugging

Versioneer tries to avoid fatal errors: if something goes wrong, it will tend
to return a version of "0+unknown". To investigate the problem, run `setup.py
version`, which will run the version-lookup code in a verbose mode, and will
display the full contents of `get_versions()` (including the `error` string,
which may help identify what went wrong).

## Updating Versioneer

To upgrade your project to a new release of Versioneer, do the following:

* install the new Versioneer (`pip install -U versioneer` or equivalent)
* edit `setup.cfg`, if necessary, to include any new configuration settings
  indicated by the release notes
* re-run `versioneer install` in your source tree, to replace
  `SRC/_version.py`
* commit any changed files

### Upgrading to 0.15

Starting with this version, Versioneer is configured with a `[versioneer]`
section in your `setup.cfg` file. Earlier versions required the `setup.py` to
set attributes on the `versioneer` module immediately after import. The new
version will refuse to run (raising an exception during import) until you
have provided the necessary `setup.cfg` section.

In addition, the Versioneer package provides an executable named
`versioneer`, and the installation process is driven by running `versioneer
install`. In 0.14 and earlier, the executable was named
`versioneer-installer` and was run without an argument.

### Upgrading to 0.14

0.14 changes the format of the version string. 0.13 and earlier used
hyphen-separated strings like "0.11-2-g1076c97-dirty". 0.14 and beyond use a
plus-separated "local version" section strings, with dot-separated
components, like "0.11+2.g1076c97". PEP440-strict tools did not like the old
format, but should be ok with the new one.

### Upgrading from 0.11 to 0.12

Nothing special.

### Upgrading from 0.10 to 0.11

You must add a `versioneer.VCS = "git"` to your `setup.py` before re-running
`setup.py setup_versioneer`. This will enable the use of additional
version-control systems (SVN, etc) in the future.

## Future Directions

This tool is designed to make it easily extended to other version-control
systems: all VCS-specific components are in separate directories like
src/git/ . The top-level `versioneer.py` script is assembled from these
components by running make-versioneer.py . In the future, make-versioneer.py
will take a VCS name as an argument, and will construct a version of
`versioneer.py` that is specific to the given VCS. It might also take the
configuration arguments that are currently provided manually during
installation by editing setup.py . Alternatively, it might go the other
direction and include code from all supported VCS systems, reducing the
number of intermediate scripts.


## License

To make Versioneer easier to embed, all its code is dedicated to the public
domain. The `_version.py` that it creates is also in the public domain.
Specifically, both are released under the Creative Commons "Public Domain
Dedication" license (CC0-1.0), as described in
https://creativecommons.org/publicdomain/zero/1.0/ .

"""

try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import errno
import io
import json
import os
import re
import subprocess
import sys


class VersioneerConfig:
    pass


def get_root():
    # we require that all commands are run from the project root, i.e. the
    # directory that contains setup.py, setup.cfg, and versioneer.py .
    root = os.path.realpath(os.path.abspath(os.getcwd()))
    setup_py = os.path.join(root, "setup.py")
    versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        # allow 'python path/to/setup.py COMMAND'
        root = os.path.dirname(os.path.realpath(os.path.abspath(sys.argv[0])))
        setup_py = os.path.join(root, "setup.py")
        versioneer_py = os.path.join(root, "versioneer.py")
    if not (os.path.exists(setup_py) or os.path.exists(versioneer_py)):
        err = (
            "Versioneer was unable to run the project root directory. "
            "Versioneer requires setup.py to be executed from "
            "its immediate directory (like 'python setup.py COMMAND'), "
            "or in a way that lets it use sys.argv[0] to find the root "
            "(like 'python path/to/setup.py COMMAND')."
        )
        raise VersioneerBadRootError(err)
    try:
        # Certain runtime workflows (setup.py install/develop in a setuptools
        # tree) execute all dependencies in a single python process, so
        # "versioneer" may be imported multiple times, and python's shared
        # module-import table will cache the first one. So we can't use
        # os.path.dirname(__file__), as that will find whichever
        # versioneer.py was first imported, even in later projects.
        me = os.path.realpath(os.path.abspath(__file__))
        if os.path.splitext(me)[0] != os.path.splitext(versioneer_py)[0]:
            print(
                "Warning: build in %s is using versioneer.py from %s"
                % (os.path.dirname(me), versioneer_py)
            )
    except NameError:
        pass
    return root


def get_config_from_root(root):
    # This might raise EnvironmentError (if setup.cfg is missing), or
    # configparser.NoSectionError (if it lacks a [versioneer] section), or
    # configparser.NoOptionError (if it lacks "VCS="). See the docstring at
    # the top of versioneer.py for instructions on writing your setup.cfg .
    setup_cfg = os.path.join(root, "setup.cfg")

    # TODO: find a py2 compatible solution for the configparser deprecation issues
    parser = configparser.SafeConfigParser()
    with io.open(setup_cfg, "rt", encoding="utf-8") as f:
        parser.readfp(f)

    VCS = parser.get("versioneer", "VCS")  # mandatory

    def get(parser, name):
        if parser.has_option("versioneer", name):
            return parser.get("versioneer", name)
        return None

    cfg = VersioneerConfig()
    cfg.VCS = VCS
    cfg.style = get(parser, "style") or ""
    cfg.versionfile_source = get(parser, "versionfile_source")
    cfg.versionfile_build = get(parser, "versionfile_build")
    cfg.tag_prefix = get(parser, "tag_prefix")
    if cfg.tag_prefix in ("''", '""'):
        cfg.tag_prefix = ""
    cfg.parentdir_prefix = get(parser, "parentdir_prefix")
    cfg.verbose = get(parser, "verbose")
    cfg.lookupfile = get(parser, "lookupfile")
    return cfg


class NotThisMethod(Exception):
    pass


# these dictionaries contain VCS-specific tools
LONG_VERSION_PY = {}
HANDLERS = {}


def register_vcs_handler(vcs, method):  # decorator
    def decorate(f):
        if vcs not in HANDLERS:
            HANDLERS[vcs] = {}
        HANDLERS[vcs][method] = f
        return f

    return decorate


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False):
    """Call the given command(s)."""
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            dispcmd = str([c] + args)
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen(
                [c] + args,
                cwd=cwd,
                stdout=subprocess.PIPE,
                stderr=(subprocess.PIPE if hide_stderr else None),
            )
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %s" % dispcmd)
                print(e)
            return None
    else:
        if verbose:
            print("unable to find command, tried %s" % (commands,))
        return None
    stdout = p.communicate()[0].strip()
    if sys.version_info[0] >= 3:
        stdout = stdout.decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %s (error)" % dispcmd)
        return None
    return stdout


LONG_VERSION_PY[
    "git"
] = '''# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# This file helps to compute a version number in source trees obtained from
# git-archive tarball (such as those provided by githubs download-from-tag
# feature). Distribution tarballs (built by setup.py sdist) and build
# directories (produced by setup.py build) will contain a much shorter file
# that just contains the computed version number.

# This file is released into the public domain. Generated by
# versioneer-0.15+dev (https://github.com/warner/python-versioneer)

"""Git implementation of _version.py."""

import errno
import io
import os
import re
import subprocess
import sys
import logging


def get_keywords():
    """Get the keywords needed to look up the version information."""
    # these strings will be replaced by git during git-archive.
    # setup.py/versioneer.py will grep for the variable names, so they must
    # each be defined on a line of their own. _version.py will just call
    # get_keywords().
    git_refnames = "%(DOLLAR)sFormat:%%d%(DOLLAR)s"
    git_full = "%(DOLLAR)sFormat:%%H%(DOLLAR)s"
    keywords = {"refnames": git_refnames, "full": git_full}
    return keywords


class VersioneerConfig:

    """Container for Versioneer configuration parameters."""


def get_config():
    """Create, populate and return the VersioneerConfig() object."""
    # these strings are filled in when 'setup.py versioneer' creates
    # _version.py
    cfg = VersioneerConfig()
    cfg.VCS = "git"
    cfg.style = "%(STYLE)s"
    cfg.tag_prefix = "%(TAG_PREFIX)s"
    cfg.parentdir_prefix = "%(PARENTDIR_PREFIX)s"
    cfg.versionfile_source = "%(VERSIONFILE_SOURCE)s"
    cfg.lookupfile = "%(LOOKUP_FILE)s"
    cfg.verbose = False
    return cfg


class NotThisMethod(Exception):

    """Exception raised if a method is not valid for the current scenario."""


LONG_VERSION_PY = {}
HANDLERS = {}


def register_vcs_handler(vcs, method):  # decorator
    """Decorator to mark a method as the handler for a particular VCS."""
    def decorate(f):
        """Store f in HANDLERS[vcs][method]."""
        if vcs not in HANDLERS:
            HANDLERS[vcs] = {}
        HANDLERS[vcs][method] = f
        return f
    return decorate


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False):
    """Call the given command(s)."""
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            dispcmd = str([c] + args)
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen([c] + args, cwd=cwd, stdout=subprocess.PIPE,
                                 stderr=(subprocess.PIPE if hide_stderr
                                         else None))
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %%s" %% dispcmd)
                print(e)
            return None
    else:
        if verbose:
            print("unable to find command, tried %%s" %% (commands,))
        return None
    stdout = p.communicate()[0].strip()
    if sys.version_info[0] >= 3:
        stdout = stdout.decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %%s (error)" %% dispcmd)
        return None
    return stdout


def versions_from_parentdir(parentdir_prefix, root, verbose):
    """Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes
    both the project name and a version string.
    """
    dirname = os.path.basename(root)
    if not dirname.startswith(parentdir_prefix):
        if verbose:
            print("guessing rootdir is '%%s', but '%%s' doesn't start with "
                  "prefix '%%s'" %% (root, dirname, parentdir_prefix))
        raise NotThisMethod("rootdir doesn't start with parentdir_prefix")
    return {"version": dirname[len(parentdir_prefix):],
            "full-revisionid": None,
            "dirty": False, "error": None}


@register_vcs_handler("git", "get_keywords")
def git_get_keywords(versionfile_abs):
    """Extract version information from the given file."""
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = {}
    try:
        f = io.open(versionfile_abs, "rt", encoding="utf-8")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["full"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return keywords


@register_vcs_handler("git", "keywords")
def git_versions_from_keywords(keywords, tag_prefix, verbose):
    """Get version information from git keywords."""
    if not keywords:
        raise NotThisMethod("no keywords at all, weird")
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        raise NotThisMethod("unexpanded keywords, not a git-archive tarball")
    refs = set(r.strip() for r in refnames.strip("()").split(","))
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set(r[len(TAG):] for r in refs if r.startswith(TAG))
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %%d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set(r for r in refs if re.search(r'\d', r))
        if verbose:
            print("discarding '%%s', no digits" %% ",".join(refs-tags))

    branches = [r for r in refs if not r.startswith(TAG)
                and r != "HEAD" and not r.startswith("refs/")]
    if verbose:
        print("likely branches: %%s" %% ",".join(sorted(branches)))
    branch = None
    if branches:
        branch = branches[0]

    if verbose:
        print("likely tags: %%s" %% ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking %%s" %% r)

            result = {"version": r,
                      "full-revisionid": keywords["full"].strip(),
                      "dirty": False, "error": None}
            if branch is not None:
                result["branch"] = branch
            return result
    # no suitable tags, so version is "0+unknown", but full hex is still there
    if verbose:
        print("no suitable tags, using unknown + full revision id")
    return {"version": "0+unknown",
            "full-revisionid": keywords["full"].strip(),
            "dirty": False, "error": "no suitable tags"}


@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, run_command=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    if not os.path.exists(os.path.join(root, ".git")):
        if verbose:
            print("no .git in %%s" %% root)
        raise NotThisMethod("no .git directory")

    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out = run_command(GITS, ["describe", "--tags", "--dirty",
                                      "--always", "--long",
                                      "--match", "%%s*" %% tag_prefix],
                               cwd=root)
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[:git_describe.rindex("-dirty")]

    # figure out our branch
    abbrev_ref_out = run_command(GITS,
                                 ["rev-parse", "--abbrev-ref", "HEAD"],
                                 cwd=root)
    if abbrev_ref_out is not None and abbrev_ref_out != "HEAD":
        pieces["branch"] = abbrev_ref_out.strip()

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r'^(.+)-(\d+)-g([0-9a-f]+)$', git_describe)
        if not mo:
            # unparseable. Maybe git-describe is misbehaving?
            pieces["error"] = ("unable to parse git-describe output: '%%s'"
                               %% describe_out)
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%%s' doesn't start with prefix '%%s'"
                print(fmt %% (full_tag, tag_prefix))
            pieces["error"] = ("tag '%%s' doesn't start with prefix '%%s'"
                               %% (full_tag, tag_prefix))
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix):]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out = run_command(GITS, ["rev-list", "HEAD", "--count"],
                                cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    return pieces


@register_vcs_handler("git", "parse_lookup_file")
def git_parse_lookup_file(path):
    """Parse a versioneer lookup file.

    This file allows definition of branch specific data like virtual tags or
    custom styles to use for version rendering.
    """
    if not os.path.exists(path):
        return []

    import re
    lookup = []
    with io.open(path, "rt", encoding="utf-8") as f:
        for line in f:
            if '#' in line:
                line = line[:line.index("#")]
            line = line.strip()
            if not line:
                continue

            try:
                split_line = list(map(lambda x: x.strip(), line.split()))
                if not len(split_line):
                    continue

                matcher = re.compile(split_line[0])

                if len(split_line) == 1:
                    entry = [matcher, None, None, None]
                elif len(split_line) == 2:
                    render = split_line[1]
                    entry = [matcher, render, None, None]
                elif len(split_line) == 3:
                    tag, ref_commit = split_line[1:]
                    entry = [matcher, None, tag, ref_commit]
                elif len(split_line) == 4:
                    tag, ref_commit, render = split_line[1:]
                    entry = [matcher, render, tag, ref_commit]
                else:
                    continue

                lookup.append(entry)
            except Exception:
                logging.getLogger(__name__).exception("Versioneer problem")
                break
    return lookup


@register_vcs_handler("git", "pieces_from_lookup")
def git_pieces_from_lookup(lookup, root, verbose, run_command=run_command):
    """Extract version information based on provided lookup data."""
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]

    stdout = run_command(GITS, ["rev-parse", "--abbrev-ref", "HEAD"],
                         cwd=root)
    if stdout is None:
        raise NotThisMethod("git rev-parse --abbrev-ref HEAD failed")

    current_branch = stdout.strip()
    if current_branch == "HEAD":
        raise NotThisMethod("not on a branch")

    for matcher, render, tag, ref_commit in lookup:
        if matcher.match(current_branch):
            if tag is None or ref_commit is None:
                raise NotThisMethod("tag or ref_commit is unset for "
                                    "this branch")

            stdout = run_command(GITS,
                                 ["rev-list", "%%s..HEAD" %% ref_commit,
                                  "--count"],
                                 cwd=root)
            if stdout is None:
                raise NotThisMethod("git rev-list %%s..HEAD "
                                    "--count failed" %% ref_commit)
            try:
                num_commits = int(stdout.strip())
            except ValueError:
                raise NotThisMethod("git rev-list %%s..HEAD --count didn't "
                                    "return a valid number" %% ref_commit)

            stdout = run_command(GITS,
                                 ["rev-parse", "--short", "HEAD"],
                                 cwd=root)
            if stdout is None:
                raise NotThisMethod("git describe rev-parse "
                                    "--short HEAD failed")
            short_hash = stdout.strip()

            stdout = run_command(GITS,
                                 ["describe", "--tags",
                                  "--dirty", "--always"],
                                 cwd=root)
            if stdout is None:
                raise NotThisMethod("git describe --tags --dirty "
                                    "--always failed")
            dirty = stdout.strip().endswith("-dirty")

            stdout = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
            if stdout is None:
                raise NotThisMethod("git rev-parse HEAD failed")
            full = stdout.strip()

            return {
                "long": full,
                "short": short_hash,
                "dirty": dirty,
                "branch": current_branch,
                "closest-tag": tag,
                "distance": num_commits,
                "error": None,
                "render": render
            }

    raise NotThisMethod("no matching lookup definition found")


def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%%d.g%%s" %% (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%%d.g%%s" %% (pieces["distance"],
                                          pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_tag(pieces):
    """TAG[[.postDISTANCE].dev0+gHEX] -- Just the tag if not dirty, else more info

    Useful for projects that want commit based tracking on some branches
    but have the master branch only report tags, to allow for commits that
    do not modify actual code (e.g. to .github/* or docs).

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]+gHEX
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["dirty"]:
            rendered += ".post%%d" %% pieces["distance"]
            rendered += ".dev0"
            rendered += "+g%%s" %% pieces["short"]
    else:
        # exception #1
        rendered = "0.post%%d" %% pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%%s" %% pieces["short"]
    return rendered


def render_pep440_pre(pieces):
    """TAG[.post.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post.devDISTANCE
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".post.dev%%d" %% pieces["distance"]
    else:
        # exception #1
        rendered = "0.post.dev%%d" %% pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%%d" %% pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%%s" %% pieces["short"]
    else:
        # exception #1
        rendered = "0.post%%d" %% pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%%s" %% pieces["short"]
    return rendered


def render_pep440_dev(pieces):
    """TAG[.devDISTANCE]+gHEX[.dirty] .

    Exceptions:
    1: no tags. 0.devDISTANCE+gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".dev%%d" %% pieces["distance"]
        rendered += plus_or_dot(pieces)
    else:
        # exception #1
        rendered = "0.dev%%d" %% pieces["distance"]
        rendered += "+"
    rendered += "g%%s" %% pieces["short"]
    if pieces["dirty"]:
        rendered += ".dirty"
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Eexceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%%d" %% pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%%d" %% pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%%d-g%%s" %% (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%%d-g%%s" %% (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {"version": "unknown",
                "full-revisionid": pieces.get("long"),
                "dirty": None,
                "error": pieces["error"]}

    if "render" in pieces and pieces["render"] is not None:
        style = pieces["render"]

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "pep440-dev":
        rendered = render_pep440_dev(pieces)
    elif style == "pep440-tag":
        rendered = render_pep440_tag(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%%s'" %% style)

    result = {"version": rendered, "full-revisionid": pieces["long"],
              "dirty": pieces["dirty"], "error": None}
    if "branch" in pieces and pieces["branch"] is not None:
        result["branch"] = pieces["branch"]
    return result


def get_versions():
    """Get version information or return default if unable to do so."""
    # I am in _version.py, which lives at ROOT/VERSIONFILE_SOURCE. If we have
    # __file__, we can work backwards from there to the root. Some
    # py2exe/bbfreeze/non-CPython implementations don't do __file__, in which
    # case we can only use expanded keywords.

    cfg = get_config()
    verbose = cfg.verbose

    try:
        return git_versions_from_keywords(get_keywords(), cfg.tag_prefix,
                                          verbose)
    except NotThisMethod:
        pass

    try:
        root = os.path.realpath(__file__)
        # versionfile_source is the relative path from the top of the source
        # tree (where the .git directory might live) to this file. Invert
        # this to find the root from __file__.
        for i in cfg.versionfile_source.split('/'):
            root = os.path.dirname(root)
    except NameError:
        return {"version": "0+unknown", "full-revisionid": None,
                "dirty": None,
                "error": "unable to find root of source tree"}

    lookupfile = cfg.lookupfile if cfg.lookupfile is not None \
        else ".versioneer-lookup"
    lookuppath = os.path.join(root, lookupfile)
    if os.path.exists(lookuppath):
        try:
            lookup_data = git_parse_lookup_file(lookuppath)
            pieces = git_pieces_from_lookup(lookup_data, root, verbose)
            return render(pieces, cfg.style)
        except NotThisMethod:
            pass

    try:
        pieces = git_pieces_from_vcs(cfg.tag_prefix, root, verbose)
        return render(pieces, cfg.style)
    except NotThisMethod:
        pass

    try:
        if cfg.parentdir_prefix:
            return versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
    except NotThisMethod:
        pass

    return {"version": "0+unknown", "full-revisionid": None,
            "dirty": None,
            "error": "unable to compute version"}
'''


@register_vcs_handler("git", "get_keywords")
def git_get_keywords(versionfile_abs):
    """Extract version information from the given file."""
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = {}
    try:
        f = io.open(versionfile_abs, "rt", encoding="utf-8")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["full"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return keywords


@register_vcs_handler("git", "keywords")
def git_versions_from_keywords(keywords, tag_prefix, verbose):
    """Get version information from git keywords."""
    if not keywords:
        raise NotThisMethod("no keywords at all, weird")
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        raise NotThisMethod("unexpanded keywords, not a git-archive tarball")
    refs = {r.strip() for r in refnames.strip("()").split(",")}
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = {r[len(TAG) :] for r in refs if r.startswith(TAG)}
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = {r for r in refs if re.search(r"\d", r)}
        if verbose:
            print("discarding '%s', no digits" % ",".join(refs - tags))

    branches = [
        r
        for r in refs
        if not r.startswith(TAG) and r != "HEAD" and not r.startswith("refs/")
    ]
    if verbose:
        print("likely branches: %s" % ",".join(sorted(branches)))
    branch = None
    if branches:
        branch = branches[0]

    if verbose:
        print("likely tags: %s" % ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix) :]
            if verbose:
                print("picking %s" % r)

            result = {
                "version": r,
                "full-revisionid": keywords["full"].strip(),
                "dirty": False,
                "error": None,
            }
            if branch is not None:
                result["branch"] = branch
            return result
    # no suitable tags, so version is "0+unknown", but full hex is still there
    if verbose:
        print("no suitable tags, using unknown + full revision id")
    return {
        "version": "0+unknown",
        "full-revisionid": keywords["full"].strip(),
        "dirty": False,
        "error": "no suitable tags",
    }


@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, run_command=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    if not os.path.exists(os.path.join(root, ".git")):
        if verbose:
            print("no .git in %s" % root)
        raise NotThisMethod("no .git directory")

    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out = run_command(
        GITS,
        [
            "describe",
            "--tags",
            "--dirty",
            "--always",
            "--long",
            "--match",
            "%s*" % tag_prefix,
        ],
        cwd=root,
    )
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[: git_describe.rindex("-dirty")]

    # figure out our branch
    abbrev_ref_out = run_command(GITS, ["rev-parse", "--abbrev-ref", "HEAD"], cwd=root)
    if abbrev_ref_out is not None and abbrev_ref_out != "HEAD":
        pieces["branch"] = abbrev_ref_out.strip()

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r"^(.+)-(\d+)-g([0-9a-f]+)$", git_describe)
        if not mo:
            # unparseable. Maybe git-describe is misbehaving?
            pieces["error"] = "unable to parse git-describe output: '%s'" % describe_out
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%s' doesn't start with prefix '%s'"
                print(fmt % (full_tag, tag_prefix))
            pieces["error"] = "tag '%s' doesn't start with prefix '%s'" % (
                full_tag,
                tag_prefix,
            )
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix) :]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out = run_command(GITS, ["rev-list", "HEAD", "--count"], cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    return pieces


@register_vcs_handler("git", "parse_lookup_file")
def git_parse_lookup_file(path):
    """Parse a versioneer lookup file.

    This file allows definition of branch specific data like virtual tags or
    custom styles to use for version rendering.
    """
    if not os.path.exists(path):
        return []

    import re

    lookup = []
    with io.open(path, "r") as f:
        for line in f:
            if "#" in line:
                line = line[: line.index("#")]
            line = line.strip()
            if not line:
                continue

            try:
                split_line = list(map(lambda x: x.strip(), line.split()))
                if not len(split_line):
                    continue

                matcher = re.compile(split_line[0])

                if len(split_line) == 1:
                    entry = [matcher, None, None, None]
                elif len(split_line) == 2:
                    render = split_line[1]
                    entry = [matcher, render, None, None]
                elif len(split_line) == 3:
                    tag, ref_commit = split_line[1:]
                    entry = [matcher, None, tag, ref_commit]
                elif len(split_line) == 4:
                    tag, ref_commit, render = split_line[1:]
                    entry = [matcher, render, tag, ref_commit]
                else:
                    continue

                lookup.append(entry)
            except Exception:
                break
    return lookup


@register_vcs_handler("git", "pieces_from_lookup")
def git_pieces_from_lookup(lookup, root, verbose, run_command=run_command):
    """Extract version information based on provided lookup data."""
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]

    stdout = run_command(GITS, ["rev-parse", "--abbrev-ref", "HEAD"], cwd=root)
    if stdout is None:
        raise NotThisMethod("git rev-parse --abbrev-ref HEAD failed")

    current_branch = stdout.strip()
    if current_branch == "HEAD":
        raise NotThisMethod("not on a branch")

    for matcher, render, tag, ref_commit in lookup:
        if matcher.match(current_branch):
            if tag is None or ref_commit is None:
                raise NotThisMethod("tag or ref_commit is unset for " "this branch")

            stdout = run_command(
                GITS, ["rev-list", "%s..HEAD" % ref_commit, "--count"], cwd=root
            )
            if stdout is None:
                raise NotThisMethod(
                    "git rev-list %s..HEAD " "--count failed" % ref_commit
                )
            try:
                num_commits = int(stdout.strip())
            except ValueError:
                raise NotThisMethod(
                    "git rev-list %s..HEAD --count didn't "
                    "return a valid number" % ref_commit
                )

            stdout = run_command(GITS, ["rev-parse", "--short", "HEAD"], cwd=root)
            if stdout is None:
                raise NotThisMethod("git describe rev-parse " "--short HEAD failed")
            short_hash = stdout.strip()

            stdout = run_command(
                GITS, ["describe", "--tags", "--dirty", "--always"], cwd=root
            )
            if stdout is None:
                raise NotThisMethod("git describe --tags --dirty " "--always failed")
            dirty = stdout.strip().endswith("-dirty")

            stdout = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
            if stdout is None:
                raise NotThisMethod("git rev-parse HEAD failed")
            full = stdout.strip()

            return {
                "long": full,
                "short": short_hash,
                "dirty": dirty,
                "branch": current_branch,
                "closest-tag": tag,
                "distance": num_commits,
                "error": None,
                "render": render,
            }

    raise NotThisMethod("no matching lookup definition found")


def do_vcs_install(manifest_in, versionfile_source, ipy):
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    files = [manifest_in, versionfile_source]
    if ipy:
        files.append(ipy)
    try:
        me = __file__
        if me.endswith(".pyc") or me.endswith(".pyo"):
            me = os.path.splitext(me)[0] + ".py"
        versioneer_file = os.path.relpath(me)
    except NameError:
        versioneer_file = "versioneer.py"
    files.append(versioneer_file)
    present = False
    try:
        f = io.open(".gitattributes", "rt", encoding="utf-8")
        for line in f.readlines():
            if line.strip().startswith(versionfile_source):
                if "export-subst" in line.strip().split()[1:]:
                    present = True
        f.close()
    except EnvironmentError:
        pass
    if not present:
        f = io.open(".gitattributes", "a+t", encoding="utf-8")
        f.write("%s export-subst\n" % versionfile_source)
        f.close()
        files.append(".gitattributes")
    run_command(GITS, ["add", "--"] + files)


def versions_from_parentdir(parentdir_prefix, root, verbose):
    """Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes
    both the project name and a version string.
    """
    dirname = os.path.basename(root)
    if not dirname.startswith(parentdir_prefix):
        if verbose:
            print(
                "guessing rootdir is '%s', but '%s' doesn't start with "
                "prefix '%s'" % (root, dirname, parentdir_prefix)
            )
        raise NotThisMethod("rootdir doesn't start with parentdir_prefix")
    return {
        "version": dirname[len(parentdir_prefix) :],
        "full-revisionid": None,
        "dirty": False,
        "error": None,
    }


SHORT_VERSION_PY = """
# This file was generated by 'versioneer.py' (0.15+dev) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json
import sys

version_json = '''
%s
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
"""


def versions_from_file(filename):
    try:
        with io.open(filename, "rt", encoding="utf-8") as f:
            contents = f.read()
    except EnvironmentError:
        raise NotThisMethod("unable to read _version.py")
    mo = re.search(
        r"version_json = '''\s+(.*)'''  # END VERSION_JSON", contents, re.M | re.S
    )
    if not mo:
        raise NotThisMethod("no version_json in _version.py")
    return json.loads(mo.group(1))


def write_to_version_file(filename, versions):
    os.unlink(filename)
    contents = json.dumps(versions, sort_keys=True, indent=1, separators=(",", ": "))
    with io.open(filename, "wt", encoding="utf-8") as f:
        f.write(SHORT_VERSION_PY % contents)

    print("set %s to '%s'" % (filename, versions["version"]))


def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%d.g%s" % (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%d.g%s" % (pieces["distance"], pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_tag(pieces):
    """TAG[[.postDISTANCE].dev0+gHEX] -- Just the tag if not dirty, else more info

    Useful for projects that want commit based tracking on some branches
    but have the master branch only report tags, to allow for commits that
    do not modify actual code (e.g. to .github/* or docs).

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]+gHEX
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            rendered += ".dev0"
            rendered += "+g%s" % pieces["short"]
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
    return rendered


def render_pep440_pre(pieces):
    """TAG[.post.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post.devDISTANCE
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".post.dev%d" % pieces["distance"]
    else:
        # exception #1
        rendered = "0.post.dev%d" % pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%s" % pieces["short"]
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
    return rendered


def render_pep440_dev(pieces):
    """TAG[.devDISTANCE]+gHEX[.dirty] .

    Exceptions:
    1: no tags. 0.devDISTANCE+gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".dev%d" % pieces["distance"]
        rendered += plus_or_dot(pieces)
    else:
        # exception #1
        rendered = "0.dev%d" % pieces["distance"]
        rendered += "+"
    rendered += "g%s" % pieces["short"]
    if pieces["dirty"]:
        rendered += ".dirty"
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Eexceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {
            "version": "unknown",
            "full-revisionid": pieces.get("long"),
            "dirty": None,
            "error": pieces["error"],
        }

    if "render" in pieces and pieces["render"] is not None:
        style = pieces["render"]

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "pep440-dev":
        rendered = render_pep440_dev(pieces)
    elif style == "pep440-tag":
        rendered = render_pep440_tag(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%s'" % style)

    result = {
        "version": rendered,
        "full-revisionid": pieces["long"],
        "dirty": pieces["dirty"],
        "error": None,
    }
    if "branch" in pieces and pieces["branch"] is not None:
        result["branch"] = pieces["branch"]
    return result


class VersioneerBadRootError(Exception):
    pass


def get_versions(verbose=False):
    # returns dict with two keys: 'version' and 'full'

    if "versioneer" in sys.modules:
        # see the discussion in cmdclass.py:get_cmdclass()
        del sys.modules["versioneer"]

    root = get_root()
    cfg = get_config_from_root(root)

    assert cfg.VCS is not None, "please set [versioneer]VCS= in setup.cfg"
    handlers = HANDLERS.get(cfg.VCS)
    assert handlers, "unrecognized VCS '%s'" % cfg.VCS
    verbose = verbose or cfg.verbose
    assert cfg.versionfile_source is not None, "please set versioneer.versionfile_source"
    assert cfg.tag_prefix is not None, "please set versioneer.tag_prefix"

    versionfile_abs = os.path.join(root, cfg.versionfile_source)

    # extract version from first of: _version.py, VCS command (e.g. 'git
    # describe'), parentdir. This is meant to work for developers using a
    # source checkout, for users of a tarball created by 'setup.py sdist',
    # and for users of a tarball/zipball created by 'git archive' or github's
    # download-from-tag feature or the equivalent in other VCSes.

    get_keywords_f = handlers.get("get_keywords")
    from_keywords_f = handlers.get("keywords")
    if get_keywords_f and from_keywords_f:
        try:
            keywords = get_keywords_f(versionfile_abs)
            ver = from_keywords_f(keywords, cfg.tag_prefix, verbose)
            if verbose:
                print("got version from expanded keyword %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        ver = versions_from_file(versionfile_abs)
        if verbose:
            print("got version from file %s %s" % (versionfile_abs, ver))
        return ver
    except NotThisMethod:
        pass

    lookupfile = cfg.lookupfile if cfg.lookupfile is not None else ".versioneer-lookup"
    lookuppath = os.path.join(root, lookupfile)
    if os.path.exists(lookuppath):
        parse_lookup_file_f = handlers.get("parse_lookup_file")
        versions_from_lookup_f = handlers.get("pieces_from_lookup")
        if parse_lookup_file_f and versions_from_lookup_f:
            try:
                lookup_data = parse_lookup_file_f(lookuppath)
                pieces = versions_from_lookup_f(lookup_data, root, verbose)
                ver = render(pieces, cfg.style)
                if verbose:
                    print("got version from lookup file")
                return ver
            except NotThisMethod:
                pass
    elif verbose:
        print("lookup file %s doesn't exist")

    from_vcs_f = handlers.get("pieces_from_vcs")
    if from_vcs_f:
        try:
            pieces = from_vcs_f(cfg.tag_prefix, root, verbose)
            ver = render(pieces, cfg.style)
            if verbose:
                print("got version from VCS %s" % ver)
            return ver
        except NotThisMethod:
            pass

    try:
        if cfg.parentdir_prefix:
            ver = versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
            if verbose:
                print("got version from parentdir %s" % ver)
            return ver
    except NotThisMethod:
        pass

    if verbose:
        print("unable to compute version")

    return {
        "version": "0+unknown",
        "full-revisionid": None,
        "dirty": None,
        "error": "unable to compute version",
    }


def get_version():
    return get_versions()["version"]


def get_cmdclass():
    if "versioneer" in sys.modules:
        del sys.modules["versioneer"]
        # this fixes the "python setup.py develop" case (also 'install' and
        # 'easy_install .'), in which subdependencies of the main project are
        # built (using setup.py bdist_egg) in the same python process. Assume
        # a main project A and a dependency B, which use different versions
        # of Versioneer. A's setup.py imports A's Versioneer, leaving it in
        # sys.modules by the time B's setup.py is executed, causing B to run
        # with the wrong versioneer. Setuptools wraps the sub-dep builds in a
        # sandbox that restores sys.modules to it's pre-build state, so the
        # parent is protected against the child's "import versioneer". By
        # removing ourselves from sys.modules here, before the child build
        # happens, we protect the child from the parent's versioneer too.
        # Also see https://github.com/warner/python-versioneer/issues/52

    cmds = {}

    # we add "version" to both distutils and setuptools
    from distutils.core import Command

    class cmd_version(Command):
        description = "report generated version string"
        user_options = []
        boolean_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            vers = get_versions(verbose=True)
            print("Version: %s" % vers["version"])
            print(" full-revisionid: %s" % vers.get("full-revisionid"))
            if "branch" in vers:
                print(" branch: %s" % vers["branch"])
            print(" dirty: %s" % vers.get("dirty"))
            if vers["error"]:
                print(" error: %s" % vers["error"])

    cmds["version"] = cmd_version

    # we override "build_py" in both distutils and setuptools
    #
    # most invocation pathways end up running build_py:
    #  distutils/build -> build_py
    #  distutils/install -> distutils/build ->..
    #  setuptools/bdist_wheel -> distutils/install ->..
    #  setuptools/bdist_egg -> distutils/install_lib -> build_py
    #  setuptools/install -> bdist_egg ->..
    #  setuptools/develop -> ?

    from distutils.command.build_py import build_py as _build_py

    class cmd_build_py(_build_py):
        def run(self):
            root = get_root()
            cfg = get_config_from_root(root)
            versions = get_versions()
            _build_py.run(self)
            # now locate _version.py in the new build/ directory and replace
            # it with an updated value
            if cfg.versionfile_build:
                target_versionfile = os.path.join(self.build_lib, cfg.versionfile_build)
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

    cmds["build_py"] = cmd_build_py

    if "cx_Freeze" in sys.modules:  # cx_freeze enabled?
        from cx_Freeze.dist import build_exe as _build_exe

        class cmd_build_exe(_build_exe):
            def run(self):
                root = get_root()
                cfg = get_config_from_root(root)
                versions = get_versions()
                target_versionfile = cfg.versionfile_source
                print("UPDATING %s" % target_versionfile)
                write_to_version_file(target_versionfile, versions)

                _build_exe.run(self)
                os.unlink(target_versionfile)
                with io.open(cfg.versionfile_source, "wt", encoding="utf-8") as f:
                    LONG = LONG_VERSION_PY[cfg.VCS]
                    f.write(
                        LONG
                        % {
                            "DOLLAR": "$",
                            "STYLE": cfg.style,
                            "TAG_PREFIX": cfg.tag_prefix,
                            "PARENTDIR_PREFIX": cfg.parentdir_prefix,
                            "VERSIONFILE_SOURCE": cfg.versionfile_source,
                        }
                    )

        cmds["build_exe"] = cmd_build_exe
        del cmds["build_py"]

    # we override different "sdist" commands for both environments
    if "setuptools" in sys.modules:
        from setuptools.command.sdist import sdist as _sdist
    else:
        from distutils.command.sdist import sdist as _sdist

    class cmd_sdist(_sdist):
        def run(self):
            versions = get_versions()
            self._versioneer_generated_versions = versions
            # unless we update this, the command will keep using the old
            # version
            self.distribution.metadata.version = versions["version"]
            return _sdist.run(self)

        def make_release_tree(self, base_dir, files):
            root = get_root()
            cfg = get_config_from_root(root)
            _sdist.make_release_tree(self, base_dir, files)
            # now locate _version.py in the new base_dir directory
            # (remembering that it may be a hardlink) and replace it with an
            # updated value
            target_versionfile = os.path.join(base_dir, cfg.versionfile_source)
            print("UPDATING %s" % target_versionfile)
            write_to_version_file(target_versionfile, self._versioneer_generated_versions)

    cmds["sdist"] = cmd_sdist

    return cmds


CONFIG_ERROR = """
setup.cfg is missing the necessary Versioneer configuration. You need
a section like:

 [versioneer]
 VCS = git
 style = pep440
 versionfile_source = src/myproject/_version.py
 versionfile_build = myproject/_version.py
 tag_prefix =
 parentdir_prefix = myproject-

You will also need to edit your setup.py to use the results:

 import versioneer
 setup(version=versioneer.get_version(),
       cmdclass=versioneer.get_cmdclass(), ...)

Please read the docstring in ./versioneer.py for configuration instructions,
edit setup.cfg, and re-run the installer or 'python versioneer.py setup'.
"""

SAMPLE_CONFIG = """
# See the docstring in versioneer.py for instructions. Note that you must
# re-run 'versioneer.py setup' after changing this section, and commit the
# resulting files.

[versioneer]
#VCS = git
#style = pep440
#versionfile_source =
#versionfile_build =
#tag_prefix =
#parentdir_prefix =
#lookupfile =

"""

INIT_PY_SNIPPET = """
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
"""


def do_setup():
    root = get_root()
    try:
        cfg = get_config_from_root(root)
    except (
        EnvironmentError,
        configparser.NoSectionError,
        configparser.NoOptionError,
    ) as e:
        if isinstance(e, (EnvironmentError, configparser.NoSectionError)):
            print("Adding sample versioneer config to setup.cfg", file=sys.stderr)
            with io.open(os.path.join(root, "setup.cfg"), "at", encoding="utf-8") as f:
                f.write(SAMPLE_CONFIG)
        print(CONFIG_ERROR, file=sys.stderr)
        return 1

    print(" creating %s" % cfg.versionfile_source)
    with io.open(cfg.versionfile_source, "wt", encoding="utf-8") as f:
        LONG = LONG_VERSION_PY[cfg.VCS]
        f.write(
            LONG
            % {
                "DOLLAR": "$",
                "STYLE": cfg.style,
                "TAG_PREFIX": cfg.tag_prefix,
                "PARENTDIR_PREFIX": cfg.parentdir_prefix,
                "VERSIONFILE_SOURCE": cfg.versionfile_source,
                "LOOKUP_FILE": cfg.lookupfile,
            }
        )

    ipy = os.path.join(os.path.dirname(cfg.versionfile_source), "__init__.py")
    if os.path.exists(ipy):
        try:
            with io.open(ipy, "rt", encoding="utf-8") as f:
                old = f.read()
        except EnvironmentError:
            old = ""
        if "from ._version import get_versions" not in old:
            print(" appending to %s" % ipy)
            with io.open(ipy, "at", encoding="utf-8") as f:
                f.write(INIT_PY_SNIPPET)
        else:
            print(" %s unmodified" % ipy)
    else:
        print(" %s doesn't exist, ok" % ipy)
        ipy = None

    # Make sure both the top-level "versioneer.py" and versionfile_source
    # (PKG/_version.py, used by runtime code) are in MANIFEST.in, so
    # they'll be copied into source distributions. Pip won't be able to
    # install the package without this.
    manifest_in = os.path.join(root, "MANIFEST.in")
    simple_includes = set()
    try:
        with io.open(manifest_in, "r") as f:
            for line in f:
                if line.startswith("include "):
                    for include in line.split()[1:]:
                        simple_includes.add(include)
    except EnvironmentError:
        pass
    # That doesn't cover everything MANIFEST.in can do
    # (http://docs.python.org/2/distutils/sourcedist.html#commands), so
    # it might give some false negatives. Appending redundant 'include'
    # lines is safe, though.
    if "versioneer.py" not in simple_includes:
        print(" appending 'versioneer.py' to MANIFEST.in")
        with io.open(manifest_in, "at", encoding="utf-8") as f:
            f.write("include versioneer.py\n")
    else:
        print(" 'versioneer.py' already in MANIFEST.in")
    if cfg.versionfile_source not in simple_includes:
        print(
            " appending versionfile_source ('%s') to MANIFEST.in" % cfg.versionfile_source
        )
        with io.open(manifest_in, "at", encoding="utf-8") as f:
            f.write("include %s\n" % cfg.versionfile_source)
    else:
        print(" versionfile_source already in MANIFEST.in")

    # Make VCS-specific changes. For git, this means creating/changing
    # .gitattributes to mark _version.py for export-time keyword
    # substitution.
    do_vcs_install(manifest_in, cfg.versionfile_source, ipy)
    return 0


def scan_setup_py():
    found = set()
    setters = False
    errors = 0
    with io.open("setup.py", "rt", encoding="utf-8") as f:
        for line in f.readlines():
            if "import versioneer" in line:
                found.add("import")
            if "versioneer.get_cmdclass()" in line:
                found.add("cmdclass")
            if "versioneer.get_version()" in line:
                found.add("get_version")
            if "versioneer.VCS" in line:
                setters = True
            if "versioneer.versionfile_source" in line:
                setters = True
    if len(found) != 3:
        print("")
        print("Your setup.py appears to be missing some important items")
        print("(but I might be wrong). Please make sure it has something")
        print("roughly like the following:")
        print("")
        print(" import versioneer")
        print(" setup( version=versioneer.get_version(),")
        print("        cmdclass=versioneer.get_cmdclass(),  ...)")
        print("")
        errors += 1
    if setters:
        print("You should remove lines like 'versioneer.VCS = ' and")
        print("'versioneer.versionfile_source = ' . This configuration")
        print("now lives in setup.cfg, and should be removed from setup.py")
        print("")
        errors += 1
    return errors


if __name__ == "__main__":
    cmd = sys.argv[1]
    if cmd == "setup":
        errors = do_setup()
        errors += scan_setup_py()
        if errors:
            sys.exit(1)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import os
import sys
from distutils.command.build_py import build_py as _build_py

import versioneer  # noqa: F401

sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), "src"))
import setuptools  # noqa: F401,E402

import octoprint_setuptools  # noqa: F401,E402

# ----------------------------------------------------------------------------------------

# Supported python versions
# we test against 2.7, 3.6 and 3.7, so that's what we'll mark as supported
PYTHON_REQUIRES = ">=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4"

# Requirements for setup.py
SETUP_REQUIRES = ["markdown>=3.1,<3.2"]  # newer versions require Python 3

# Requirements for our application
INSTALL_REQUIRES = [
    # additional OctoPrint plugins that are maintained on a different release cycle
    "OctoPrint-FirmwareCheck>=2020.09.23",
    "OctoPrint-FileCheck>=2020.08.07",
    # the following dependencies are non trivial to update since later versions
    # introduce backwards incompatible changes that might affect plugins, or due to
    # other observed problems
    "markupsafe>=1.1,<2.0",  # Jinja dependency, newer versions require Python 3
    "tornado==5.1.1",  # newer versions require Python 3
    "markdown>=3.1,<3.2",  # newer versions require Python 3
    "rsa==4.0",  # newer versions require Python 3
    "regex!=2018.11.6",  # avoid broken 2018.11.6. See #2874
    # anything below this should be checked on releases for new versions
    "flask>=1.1.2,<2",
    "Jinja2>=2.11.2,<3",
    "Flask-Login>=0.5,<0.6",  # flask-login doesn't use semver & breaks stuff on minor version increases
    "Flask-Babel>=1.0,<2",
    "Flask-Assets>=2.0,<3",
    "werkzeug>=1.0.1,<2",
    "itsdangerous>=1.1.0,<2",
    "cachelib>=0.1,<1",
    "PyYAML>=5.3.1,<6",
    "pyserial>=3.4,<4",
    "netaddr>=0.7.19,<1",
    "watchdog==0.10.4",  # watchdog dropped Py>3.6 on 0.10.5 point release...
    "sarge==0.1.5post0",
    "netifaces>=0.10.9,<1",
    "pylru>=1.2,<2",
    "pkginfo>=1.5.0.1,<2",
    "requests>=2.23.0,<3",
    "semantic_version>=2.8.5,<3",
    "psutil>=5.7,<6",
    "Click>=7.1.2,<8",
    "future>=0.18.2,<1",
    "websocket-client>=0.57,<1",
    "wrapt>=1.12.1,<2",
    "emoji>=0.5.4,<1",
    "frozendict>=1.2,<2",
    "sentry-sdk>=0.15.1,<1",
    "filetype>=1.0.7,<2",
    # vendor bundled dependencies
    "unidecode>=0.04.14,<0.05",  # dependency of awesome-slugify
    "blinker>=1.4,<2",  # dependency of flask_principal
]

# Python 2 specific requirements
INSTALL_REQUIRES_PYTHON2 = [
    "feedparser>=5.2.1,<6",  # newer versions require Python 3
    "futures>=3.3,<4",
    "monotonic>=1.5,<2",
    "scandir>=1.10,<2",
    "chainmap>=1.0.3,<2",
    "typing>=3.7.4.1,<4",
    "enum34>=1.1.10,<1.2",
]

# Python 3 specific requirements
INSTALL_REQUIRES_PYTHON3 = ["feedparser>=6.0.2,<7", "zeroconf>=0.24,<0.25"]

# OSX specific requirements
INSTALL_REQUIRES_OSX = [
    "appdirs>=1.4.4",
]

# Additional requirements for optional install options
EXTRA_REQUIRES = {
    "develop": [
        # Testing dependencies
        "mock>=3.0.5,<4",
        "pytest==4.6.10",
        "pytest-doctest-custom>=1.0.0,<2",
        "ddt",
        # pre-commit
        "pre-commit",
        # profiler
        "pyinstrument",
    ],
    # Dependencies for developing OctoPrint plugins
    "plugins": ["cookiecutter>=1.7.2,<1.8"],
    # Dependencies for building the documentation - Python 3 required!
    "docs": [
        "sphinx>=3,<4",
        "sphinxcontrib-httpdomain",
        "sphinxcontrib-mermaid",
        "sphinx_rtd_theme",
        "readthedocs-sphinx-ext",
    ],
}

# Dependency links for any of the aforementioned dependencies
DEPENDENCY_LINKS = []

# adapted from https://hynek.me/articles/conditional-python-dependencies/
if int(setuptools.__version__.split(".", 1)[0]) < 18:
    # no bdist_wheel support for setuptools < 18 since we build universal wheels and our optional dependencies
    # would get lost there
    assert "bdist_wheel" not in sys.argv

    # add optional dependencies for setuptools versions < 18 that don't yet support environment markers
    if sys.version_info[0] < 3:
        INSTALL_REQUIRES += INSTALL_REQUIRES_PYTHON2
    else:
        INSTALL_REQUIRES += INSTALL_REQUIRES_PYTHON3

    if sys.platform == "darwin":
        INSTALL_REQUIRES += INSTALL_REQUIRES_OSX
else:
    # environment markers supported
    EXTRA_REQUIRES[":python_version < '3'"] = INSTALL_REQUIRES_PYTHON2
    EXTRA_REQUIRES[":python_version >= '3'"] = INSTALL_REQUIRES_PYTHON3
    EXTRA_REQUIRES[":sys_platform == 'darwin'"] = INSTALL_REQUIRES_OSX

# -----------------------------------------------------------------------------------------------------------------------
# Anything below here is just command setup and general setup configuration

here = os.path.abspath(os.path.dirname(__file__))


def read_file_contents(path):
    import codecs

    with codecs.open(path, encoding="utf-8") as f:
        return f.read()


def md_to_html_build_py_factory(files, baseclass):
    class md_to_html_build_py(baseclass):
        files = {}

        def run(self):
            print("RUNNING md_to_html_build_py")
            if not self.dry_run:
                for directory, files in self.files.items():
                    target_dir = os.path.join(self.build_lib, directory)
                    self.mkpath(target_dir)

                    for entry in files:
                        if isinstance(entry, tuple):
                            if len(entry) != 2:
                                continue
                            source, dest = entry[0], os.path.join(target_dir, entry[1])
                        else:
                            source = entry
                            dest = os.path.join(target_dir, source + ".html")

                        print("Rendering markdown from {} to {}".format(source, dest))

                        from markdown import markdownFromFile

                        markdownFromFile(input=source, output=dest, encoding="utf-8")
            baseclass.run(self)

    return type(md_to_html_build_py)(
        md_to_html_build_py.__name__, (md_to_html_build_py,), {"files": files}
    )


def get_cmdclass():
    # make sure these are always available, even when run by dependabot
    global versioneer, octoprint_setuptools, md_to_html_build_py_factory

    cmdclass = versioneer.get_cmdclass()

    # add clean command
    cmdclass.update(
        {
            "clean": octoprint_setuptools.CleanCommand.for_options(
                source_folder="src", eggs=["OctoPrint*.egg-info"]
            )
        }
    )

    # add translation commands
    translation_dir = "translations"
    pot_file = os.path.join(translation_dir, "messages.pot")
    bundled_dir = os.path.join("src", "octoprint", "translations")
    cmdclass.update(
        octoprint_setuptools.get_babel_commandclasses(
            pot_file=pot_file,
            output_dir=translation_dir,
            pack_name_prefix="OctoPrint-i18n-",
            pack_path_prefix="",
            bundled_dir=bundled_dir,
        )
    )

    cmdclass["build_py"] = md_to_html_build_py_factory(
        {
            "octoprint/templates/_data": [
                "AUTHORS.md",
                "SUPPORTERS.md",
                "THIRDPARTYLICENSES.md",
            ]
        },
        cmdclass["build_py"] if "build_py" in cmdclass else _build_py,
    )

    return cmdclass


def params():
    # make sure these are always available, even when run by dependabot
    global versioneer, get_cmdclass, read_file_contents, here, PYTHON_REQUIRES, SETUP_REQUIRES, INSTALL_REQUIRES, EXTRA_REQUIRES, DEPENDENCY_LINKS

    name = "OctoPrint"
    version = versioneer.get_version()
    cmdclass = get_cmdclass()

    description = "The snappy web interface for your 3D printer"
    long_description = read_file_contents(os.path.join(here, "README.md"))
    long_description_content_type = "text/markdown"

    python_requires = PYTHON_REQUIRES
    setup_requires = SETUP_REQUIRES
    install_requires = INSTALL_REQUIRES
    extras_require = EXTRA_REQUIRES
    dependency_links = DEPENDENCY_LINKS

    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Natural Language :: German",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: JavaScript",
        "Topic :: Printing",
        "Topic :: System :: Monitoring",
    ]
    author = "Gina Häußge"
    author_email = "gina@octoprint.org"
    url = "https://octoprint.org"
    license = "GNU Affero General Public License v3"
    keywords = "3dprinting 3dprinter 3d-printing 3d-printer octoprint"

    project_urls = {
        "Community Forum": "https://community.octoprint.org",
        "Bug Reports": "https://github.com/foosel/OctoPrint/issues",
        "Source": "https://github.com/foosel/OctoPrint",
        "Funding": "https://donate.octoprint.org",
    }

    packages = setuptools.find_packages(where="src")
    package_dir = {
        "": "src",
    }
    package_data = {
        "octoprint": octoprint_setuptools.package_data_dirs(
            "src/octoprint", ["static", "templates", "plugins", "translations"]
        )
        + ["util/piptestballoon/setup.py"]
    }

    include_package_data = True
    zip_safe = False

    if os.environ.get("READTHEDOCS", None) == "True":
        # we can't tell read the docs to please perform a pip install -e .[docs], so we help
        # it a bit here by explicitly adding the development dependencies, which include our
        # documentation dependencies
        install_requires = install_requires + extras_require["docs"]

    entry_points = {"console_scripts": ["octoprint = octoprint:main"]}

    return locals()


setuptools.setup(**params())



# -*- coding: utf-8 -*-
"""
This module contains a functions that monkey patch third party dependencies in the one or other way.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"

import functools
import sys
import warnings


def patch_sarge_async_on_py2():
    if sys.version_info[0] >= 3:
        return

    def move_async(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            if "async" in kwargs:
                warnings.warn(
                    'sarge\'s "async" parameter is deprecated. Use "_async" instead.',
                    DeprecationWarning,
                    stacklevel=2,
                )
                kwargs["async_"] = kwargs.pop("async")
            return f(*args, **kwargs)

        return decorated_function

    import sarge

    # wrap helper
    sarge.run = move_async(sarge.run)

    # wrap Command
    sarge.Command.run = move_async(sarge.Command.run)

    # wrap Pipeline
    sarge.Pipeline.run = move_async(sarge.Pipeline.run)
    sarge.Pipeline.run_command_node = move_async(sarge.Pipeline.run_command_node)
    sarge.Pipeline.run_list_node = move_async(sarge.Pipeline.run_list_node)
    sarge.Pipeline.run_logical_node = move_async(sarge.Pipeline.run_logical_node)
    sarge.Pipeline.run_node = move_async(sarge.Pipeline.run_node)
    sarge.Pipeline.run_node_in_thread = move_async(sarge.Pipeline.run_node_in_thread)
    sarge.Pipeline.run_pipeline_node = move_async(sarge.Pipeline.run_pipeline_node)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import collections
import logging
import threading

import pkg_resources

from octoprint.util.pip import UnknownPip, create_pip_caller
from octoprint.util.version import get_comparable_version

from .. import exceptions

logger = logging.getLogger("octoprint.plugins.softwareupdate.updaters.pip")
console_logger = logging.getLogger(
    "octoprint.plugins.softwareupdate.updaters.pip.console"
)

_ALREADY_INSTALLED = "Requirement already satisfied (use --upgrade to upgrade)"
_POTENTIAL_EGG_PROBLEM_POSIX = "No such file or directory"
_POTENTIAL_EGG_PROBLEM_WINDOWS = "The system cannot find the file specified"

_pip_callers = {}
_pip_caller_mutex = collections.defaultdict(threading.RLock)
_pip_version_dependency_links = pkg_resources.parse_version("1.5")


def can_perform_update(target, check, online=True):
    from .. import MINIMUM_PIP

    pip_caller = _get_pip_caller(
        command=check["pip_command"] if "pip_command" in check else None
    )
    return (
        "pip" in check
        and pip_caller is not None
        and pip_caller.available
        and pip_caller.version >= get_comparable_version(MINIMUM_PIP)
        and (online or check.get("offline", False))
    )


def _get_pip_caller(command=None):
    global _pip_callers
    global _pip_caller_mutex

    key = command
    if command is None:
        key = "__default"

    with _pip_caller_mutex[key]:
        if key not in _pip_callers:
            try:
                _pip_callers[key] = create_pip_caller(command=command)
            except UnknownPip:
                pass

        return _pip_callers.get(key)


def perform_update(target, check, target_version, log_cb=None, online=True, force=False):
    pip_command = check.get("pip_command")
    pip_working_directory = check.get("pip_cwd")

    if not online and not check.get("offline", False):
        raise exceptions.CannotUpdateOffline()

    pip_caller = _get_pip_caller(command=pip_command)
    if pip_caller is None:
        raise exceptions.UpdateError("Can't run pip", None)

    def _log_call(*lines):
        _log(lines, prefix=" ", stream="call")

    def _log_stdout(*lines):
        _log(lines, prefix=">", stream="stdout")

    def _log_stderr(*lines):
        _log(lines, prefix="!", stream="stderr")

    def _log_message(*lines):
        _log(lines, prefix="#", stream="message")

    def _log(lines, prefix=None, stream=None):
        if log_cb is None:
            return
        log_cb(lines, prefix=prefix, stream=stream)

    if log_cb is not None:
        pip_caller.on_log_call = _log_call
        pip_caller.on_log_stdout = _log_stdout
        pip_caller.on_log_stderr = _log_stderr

    install_arg = check["pip"].format(
        target_version=target_version, target=target_version
    )

    logger.debug("Target: %s, executing pip install %s" % (target, install_arg))
    pip_args = ["--disable-pip-version-check", "install", install_arg, "--no-cache-dir"]
    pip_kwargs = {
        "env": {"PYTHONWARNINGS": b"ignore:DEPRECATION::pip._internal.cli.base_command"}
    }
    if pip_working_directory is not None:
        pip_kwargs.update(cwd=pip_working_directory)

    if "dependency_links" in check and check["dependency_links"]:
        pip_args += ["--process-dependency-links"]

    returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
    if returncode != 0:
        # This could be caused by an issue with pip/setuptools: If something (target or dependency of target) was
        # installed as an egg at an earlier date (e.g. thanks to just running python setup.py install), pip install
        # will throw an error after updating that something to a newer (non-egg) version since it will still have
        # the egg on its sys.path and expect to read data from it. Running the exact same command a second time
        # solved this during hunt for a workaround, so this is what we'll now do if it indeed looks like this
        # specific problem
        def is_egg_problem(line):
            return (
                _POTENTIAL_EGG_PROBLEM_POSIX in line
                or _POTENTIAL_EGG_PROBLEM_WINDOWS in line
            ) and ".egg" in line

        if any(map(lambda x: is_egg_problem(x), stderr)) or any(
            map(lambda x: is_egg_problem(x), stdout)
        ):
            _log_message(
                'This looks like an error caused by a specific issue in upgrading Python "eggs"',
                "via current versions of pip.",
                "Performing a second install attempt as a work around.",
            )
            returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
            if returncode != 0:
                raise exceptions.UpdateError(
                    "Error while executing pip install", (stdout, stderr)
                )
        else:
            raise exceptions.UpdateError(
                "Error while executing pip install", (stdout, stderr)
            )

    if not force and any(
        map(
            lambda x: x.strip().startswith(_ALREADY_INSTALLED)
            and (install_arg in x or install_arg in x.lower()),
            stdout,
        )
    ):
        _log_message(
            "Looks like we were already installed in this version. Forcing a reinstall."
        )
        force = True

    if force:
        logger.debug(
            "Target: %s, executing pip install %s --ignore-reinstalled --force-reinstall --no-deps"
            % (target, install_arg)
        )
        pip_args += ["--ignore-installed", "--force-reinstall", "--no-deps"]

        returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
        if returncode != 0:
            raise exceptions.UpdateError(
                "Error while executing pip install --force-reinstall", (stdout, stderr)
            )

    return "ok"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext

import octoprint.plugin


class GcodeviewerPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SettingsPlugin,
):
    def get_assets(self):
        js = [
            "js/gcodeviewer.js",
            "js/viewer/ui.js",
            "js/viewer/reader.js",
            "js/viewer/renderer.js",
        ]

        return {
            "js": js,
            "less": ["less/gcodeviewer.less"],
            "css": ["css/gcodeviewer.css"],
        }

    def get_template_configs(self):
        return [
            {
                "type": "tab",
                "template": "gcodeviewer_tab.jinja2",
                "div": "gcode",
                "styles": ["display: none;"],
                "data_bind": "visible: loginState.hasAllPermissionsKo(access.permissions.GCODE_VIEWER, access.permissions.FILES_DOWNLOAD)",
            },
            {
                "type": "settings",
                "template": "gcodeviewer_settings.jinja2",
                "custom_bindings": True,
            },
            {"type": "generic", "template": "gcodeviewer_initscript.jinja2"},
        ]

    def get_settings_defaults(self):
        return {
            "mobileSizeThreshold": 2 * 1024 * 1024,  # 2MB
            "sizeThreshold": 20 * 1024 * 1024,
        }  # 20MB

    def get_settings_version(self):
        return 1

    def on_settings_migrate(self, target, current):
        if current is None:
            config = self._settings.global_get(["gcodeViewer"])
            if config:
                self._logger.info(
                    "Migrating settings from gcodeViewer to plugins.gcodeviewer..."
                )
                if "mobileSizeThreshold" in config:
                    self._settings.set_int(
                        ["mobileSizeThreshold"], config["mobileSizeThreshold"]
                    )
                if "sizeThreshold" in config:
                    self._settings.set_int(["sizeThreshold"], config["sizeThreshold"])
                self._settings.global_remove(["gcodeViewer"])


__plugin_name__ = gettext("GCode Viewer")
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Provides a GCODE viewer in OctoPrint's UI."
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin the GCode Viewer in OctoPrint will no longer be " "available."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=2.7,<4"
__plugin_implementation__ = GcodeviewerPlugin()
# __plugin_hooks__ = {
# 	"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
# }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import jsonify, make_response, request

from octoprint.access.permissions import Permissions
from octoprint.server import NO_CONTENT, printer, printerProfileManager
from octoprint.server.api import api
from octoprint.server.util.flask import get_json_command_from_request, no_firstrun_access
from octoprint.settings import settings


@api.route("/connection", methods=["GET"])
@Permissions.STATUS.require(403)
def connectionState():
    state, port, baudrate, printer_profile = printer.get_current_connection()
    current = {
        "state": state,
        "port": port,
        "baudrate": baudrate,
        "printerProfile": printer_profile["id"]
        if printer_profile is not None and "id" in printer_profile
        else "_default",
    }

    return jsonify({"current": current, "options": _get_options()})


@api.route("/connection", methods=["POST"])
@no_firstrun_access
@Permissions.CONNECTION.require(403)
def connectionCommand():
    valid_commands = {"connect": [], "disconnect": [], "fake_ack": []}

    command, data, response = get_json_command_from_request(request, valid_commands)
    if response is not None:
        return response

    if command == "connect":
        connection_options = printer.__class__.get_connection_options()

        port = None
        baudrate = None
        printerProfile = None
        if "port" in data:
            port = data["port"]
            if port not in connection_options["ports"] and port != "AUTO":
                return make_response("Invalid port: %s" % port, 400)
        if "baudrate" in data:
            baudrate = data["baudrate"]
            if baudrate not in connection_options["baudrates"] and baudrate != 0:
                return make_response("Invalid baudrate: %d" % baudrate, 400)
        if "printerProfile" in data:
            printerProfile = data["printerProfile"]
            if not printerProfileManager.exists(printerProfile):
                return make_response("Invalid printer profile: %s" % printerProfile, 400)
        if "save" in data and data["save"]:
            settings().set(["serial", "port"], port)
            settings().setInt(["serial", "baudrate"], baudrate)
            printerProfileManager.set_default(printerProfile)
        if "autoconnect" in data:
            settings().setBoolean(["serial", "autoconnect"], data["autoconnect"])
        settings().save()
        printer.connect(port=port, baudrate=baudrate, profile=printerProfile)
    elif command == "disconnect":
        printer.disconnect()
    elif command == "fake_ack":
        printer.fake_ack()

    return NO_CONTENT


def _get_options():
    connection_options = printer.__class__.get_connection_options()
    profile_options = printerProfileManager.get_all()
    default_profile = printerProfileManager.get_default()

    options = {
        "ports": connection_options["ports"],
        "baudrates": connection_options["baudrates"],
        "printerProfiles": [
            {
                "id": printer_profile["id"],
                "name": printer_profile["name"]
                if "name" in printer_profile
                else printer_profile["id"],
            }
            for printer_profile in profile_options.values()
            if "id" in printer_profile
        ],
        "portPreference": connection_options["portPreference"],
        "baudratePreference": connection_options["baudratePreference"],
        "printerProfilePreference": default_profile["id"]
        if "id" in default_profile
        else None,
    }

    return options



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module represents OctoPrint's plugin subsystem. This includes management and helper methods as well as the
registered plugin types.

.. autofunction:: plugin_manager

.. autofunction:: plugin_settings

.. autofunction:: call_plugin

.. autoclass:: PluginSettings
   :members:
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os

from octoprint.plugin.core import Plugin, PluginInfo, PluginManager  # noqa: F401
from octoprint.plugin.types import *  # noqa: F401,F403 ## used by multiple other modules
from octoprint.plugin.types import OctoPrintPlugin, SettingsPlugin
from octoprint.settings import settings as s
from octoprint.util import deprecated, to_native_str

# singleton
_instance = None


def _validate_plugin(phase, plugin_info):
    return True


def plugin_manager(
    init=False,
    plugin_folders=None,
    plugin_bases=None,
    plugin_entry_points=None,
    plugin_disabled_list=None,
    plugin_blacklist=None,
    plugin_restart_needing_hooks=None,
    plugin_obsolete_hooks=None,
    plugin_considered_bundled=None,
    plugin_validators=None,
    compatibility_ignored_list=None,
):
    """
    Factory method for initially constructing and consecutively retrieving the :class:`~octoprint.plugin.core.PluginManager`
    singleton.

    Arguments:
        init (boolean): A flag indicating whether this is the initial call to construct the singleton (True) or not
            (False, default). If this is set to True and the plugin manager has already been initialized, a :class:`ValueError`
            will be raised. The same will happen if the plugin manager has not yet been initialized and this is set to
            False.
        plugin_folders (list): A list of folders (as strings containing the absolute path to them) in which to look for
            potential plugin modules. If not provided this defaults to the configured ``plugins`` base folder and
            ``src/plugins`` within OctoPrint's code base.
        plugin_bases (list): A list of recognized plugin base classes for which to look for provided implementations. If not
            provided this defaults to :class:`~octoprint.plugin.OctoPrintPlugin`.
        plugin_entry_points (list): A list of entry points pointing to modules which to load as plugins. If not provided
            this defaults to the entry point ``octoprint.plugin``.
        plugin_disabled_list (list): A list of plugin identifiers that are currently disabled. If not provided this
            defaults to all plugins for which ``enabled`` is set to ``False`` in the settings.
        plugin_blacklist (list): A list of plugin identifiers/identifier-version tuples that are currently blacklisted.
        plugin_restart_needing_hooks (list): A list of hook namespaces which cause a plugin to need a restart in order
            be enabled/disabled. Does not have to contain full hook identifiers, will be matched with startswith similar
            to logging handlers
        plugin_obsolete_hooks (list): A list of hooks that have been declared obsolete. Plugins implementing them will
            not be enabled since they might depend on functionality that is no longer available.
        plugin_considered_bundled (list): A list of plugin identifiers that are considered bundled plugins even if
            installed separately.
        plugin_validators (list): A list of additional plugin validators through which to process each plugin.
        compatibility_ignored_list (list): A list of plugin keys for which it will be ignored if they are flagged as
            incompatible. This is for development purposes only and should not be used in production.

    Returns:
        PluginManager: A fully initialized :class:`~octoprint.plugin.core.PluginManager` instance to be used for plugin
            management tasks.

    Raises:
        ValueError: ``init`` was True although the plugin manager was already initialized, or it was False although
            the plugin manager was not yet initialized.
    """

    global _instance
    if _instance is not None:
        if init:
            raise ValueError("Plugin Manager already initialized")

    else:
        if init:
            if plugin_bases is None:
                plugin_bases = [OctoPrintPlugin]

            if plugin_restart_needing_hooks is None:
                plugin_restart_needing_hooks = [
                    "octoprint.server.http.*",
                    "octoprint.printer.factory",
                    "octoprint.access.permissions",
                    "octoprint.timelapse.extensions",
                ]

            if plugin_obsolete_hooks is None:
                plugin_obsolete_hooks = ["octoprint.comm.protocol.gcode"]

            if plugin_considered_bundled is None:
                plugin_considered_bundled = ["firmware_check", "file_check"]

            if plugin_validators is None:
                plugin_validators = [_validate_plugin]
            else:
                plugin_validators.append(_validate_plugin)

            _instance = PluginManager(
                plugin_folders,
                plugin_bases,
                plugin_entry_points,
                logging_prefix="octoprint.plugins.",
                plugin_disabled_list=plugin_disabled_list,
                plugin_blacklist=plugin_blacklist,
                plugin_restart_needing_hooks=plugin_restart_needing_hooks,
                plugin_obsolete_hooks=plugin_obsolete_hooks,
                plugin_considered_bundled=plugin_considered_bundled,
                plugin_validators=plugin_validators,
                compatibility_ignored_list=compatibility_ignored_list,
            )
        else:
            raise ValueError("Plugin Manager not initialized yet")
    return _instance


def plugin_settings(
    plugin_key,
    defaults=None,
    get_preprocessors=None,
    set_preprocessors=None,
    settings=None,
):
    """
    Factory method for creating a :class:`PluginSettings` instance.

    Arguments:
        plugin_key (string): The plugin identifier for which to create the settings instance.
        defaults (dict): The default settings for the plugin, if different from get_settings_defaults.
        get_preprocessors (dict): The getter preprocessors for the plugin.
        set_preprocessors (dict): The setter preprocessors for the plugin.
        settings (octoprint.settings.Settings): The settings instance to use.

    Returns:
        PluginSettings: A fully initialized :class:`PluginSettings` instance to be used to access the plugin's
            settings
    """
    if settings is None:
        settings = s()
    return PluginSettings(
        settings,
        plugin_key,
        defaults=defaults,
        get_preprocessors=get_preprocessors,
        set_preprocessors=set_preprocessors,
    )


def plugin_settings_for_settings_plugin(plugin_key, instance, settings=None):
    """
    Factory method for creating a :class:`PluginSettings` instance for a given :class:`SettingsPlugin` instance.

    Will return `None` if the provided `instance` is not a :class:`SettingsPlugin` instance.

    Arguments:
        plugin_key (string): The plugin identifier for which to create the settings instance.
        implementation (octoprint.plugin.SettingsPlugin): The :class:`SettingsPlugin` instance.
        settings (octoprint.settings.Settings): The settings instance to use. Defaults to the global OctoPrint settings.

    Returns:
        PluginSettings or None: A fully initialized :class:`PluginSettings` instance to be used to access the plugin's
            settings, or `None` if the provided `instance` was not a class:`SettingsPlugin`
    """
    if not isinstance(instance, SettingsPlugin):
        return None

    try:
        get_preprocessors, set_preprocessors = instance.get_settings_preprocessors()
    except Exception:
        logging.getLogger(__name__).exception(
            "Error while retrieving preprocessors for plugin {}".format(plugin_key)
        )
        return None

    return plugin_settings(
        plugin_key,
        get_preprocessors=get_preprocessors,
        set_preprocessors=set_preprocessors,
        settings=settings,
    )


def call_plugin(
    types,
    method,
    args=None,
    kwargs=None,
    callback=None,
    error_callback=None,
    sorting_context=None,
    initialized=True,
):
    """
    Helper method to invoke the indicated ``method`` on all registered plugin implementations implementing the
    indicated ``types``. Allows providing method arguments and registering callbacks to call in case of success
    and/or failure of each call which can be used to return individual results to the calling code.

    Example:

    .. sourcecode:: python

       def my_success_callback(name, plugin, result):
           print("{name} was called successfully and returned {result!r}".format(**locals()))

       def my_error_callback(name, plugin, exc):
           print("{name} raised an exception: {exc!s}".format(**locals()))

       octoprint.plugin.call_plugin(
           [octoprint.plugin.StartupPlugin],
           "on_startup",
           args=(my_host, my_port),
           callback=my_success_callback,
           error_callback=my_error_callback
       )

    Arguments:
        types (list): A list of plugin implementation types to match against.
        method (string): Name of the method to call on all matching implementations.
        args (tuple): A tuple containing the arguments to supply to the called ``method``. Optional.
        kwargs (dict): A dictionary containing the keyword arguments to supply to the called ``method``. Optional.
        callback (function): A callback to invoke after an implementation has been called successfully. Will be called
            with the three arguments ``name``, ``plugin`` and ``result``. ``name`` will be the plugin identifier,
            ``plugin`` the plugin implementation instance itself and ``result`` the result returned from the
            ``method`` invocation.
        error_callback (function): A callback to invoke after the call of an implementation resulted in an exception.
            Will be called with the three arguments ``name``, ``plugin`` and ``exc``. ``name`` will be the plugin
            identifier, ``plugin`` the plugin implementation instance itself and ``exc`` the caught exception.
        initialized (boolean): Ignored.
    """

    if not isinstance(types, (list, tuple)):
        types = [types]
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    logger = logging.getLogger(__name__)

    plugins = plugin_manager().get_implementations(
        *types, sorting_context=sorting_context
    )
    for plugin in plugins:
        if not hasattr(plugin, "_identifier"):
            continue

        if hasattr(plugin, method):
            logger.debug("Calling {} on {}".format(method, plugin._identifier))
            try:
                result = getattr(plugin, method)(*args, **kwargs)
                if callback:
                    callback(plugin._identifier, plugin, result)
            except Exception as exc:
                logger.exception(
                    "Error while calling plugin %s" % plugin._identifier,
                    extra={"plugin": plugin._identifier},
                )
                if error_callback:
                    error_callback(plugin._identifier, plugin, exc)


class PluginSettings(object):
    """
    The :class:`PluginSettings` class is the interface for plugins to their own or globally defined settings.

    It provides some convenience methods for directly accessing plugin settings via the regular
    :class:`octoprint.settings.Settings` interfaces as well as means to access plugin specific folder locations.

    All getter and setter methods will ensure that plugin settings are stored in their correct location within the
    settings structure by modifying the supplied paths accordingly.

    Arguments:
        settings (Settings): The :class:`~octoprint.settings.Settings` instance on which to operate.
        plugin_key (str): The plugin identifier of the plugin for which to create this instance.
        defaults (dict): The plugin's defaults settings, will be used to determine valid paths within the plugin's
            settings structure

    .. method:: get(path, merged=False, asdict=False)

       Retrieves a raw value from the settings for ``path``, optionally merging the raw value with the default settings
       if ``merged`` is set to True.

       :param path: The path for which to retrieve the value.
       :type path: list, tuple
       :param boolean merged: Whether to merge the returned result with the default settings (True) or not (False,
           default).
       :returns: The retrieved settings value.
       :rtype: object

    .. method:: get_int(path, min=None, max=None)

       Like :func:`get` but tries to convert the retrieved value to ``int``. If ``min`` is provided and the retrieved
       value is less than it, it will be returned instead of the value. Likewise for ``max`` - it will be returned if
       the value is greater than it.

    .. method:: get_float(path, min=None, max=None)

       Like :func:`get` but tries to convert the retrieved value to ``float``. If ``min`` is provided and the retrieved
       value is less than it, it will be returned instead of the value. Likewise for ``max`` - it will be returned if
       the value is greater than it.

    .. method:: get_boolean(path)

       Like :func:`get` but tries to convert the retrieved value to ``boolean``.

    .. method:: set(path, value, force=False)

       Sets the raw value on the settings for ``path``.

       :param path: The path for which to retrieve the value.
       :type path: list, tuple
       :param object value: The value to set.
       :param boolean force: If set to True, the modified configuration will even be written back to disk if
           the value didn't change.

    .. method:: set_int(path, value, force=False, min=None, max=None)

       Like :func:`set` but ensures the value is an ``int`` through attempted conversion before setting it.
       If ``min`` and/or ``max`` are provided, it will also be ensured that the value is greater than or equal
       to ``min`` and less than or equal to ``max``. If that is not the case, the limit value (``min`` if less than
       that, ``max`` if greater than that) will be set instead.

    .. method:: set_float(path, value, force=False, min=None, max=None)

       Like :func:`set` but ensures the value is an ``float`` through attempted conversion before setting it.
       If ``min`` and/or ``max`` are provided, it will also be ensured that the value is greater than or equal
       to ``min`` and less than or equal to ``max``. If that is not the case, the limit value (``min`` if less than
       that, ``max`` if greater than that) will be set instead.

    .. method:: set_boolean(path, value, force=False)

       Like :func:`set` but ensures the value is an ``boolean`` through attempted conversion before setting it.

    .. method:: save(force=False, trigger_event=False)

       Saves the settings to ``config.yaml`` if there are active changes. If ``force`` is set to ``True`` the settings
       will be saved even if there are no changes. Settings ``trigger_event`` to ``True`` will cause a ``SettingsUpdated``
       :ref:`event <sec-events-available_events-settings>` to get triggered.

       :param force: Force saving to ``config.yaml`` even if there are no changes.
       :type force: boolean
       :param trigger_event: Trigger the ``SettingsUpdated`` :ref:`event <sec-events-available_events-settings>` on save.
       :type trigger_event: boolean

    .. method:: add_overlay(overlay, at_end=False, key=None)

       Adds a new config overlay for the plugin's settings. Will return the overlay's key in the map.

       :param overlay: Overlay dict to add
       :type overlay: dict
       :param at_end: Whether to add overlay at end or start (default) of config hierarchy
       :type at_end: boolean
       :param key: Key to use to identify overlay. If not set one will be built based on the overlay's hash
       :type key: str
       :rtype: str

    .. method:: remove_overlay(key)

       Removes an overlay from the settings based on its key. Return ``True`` if the overlay could be found and was
       removed, ``False`` otherwise.

       :param key: The key of the overlay to remove
       :type key: str
       :rtype: boolean
    """

    def __init__(
        self,
        settings,
        plugin_key,
        defaults=None,
        get_preprocessors=None,
        set_preprocessors=None,
    ):
        self.settings = settings
        self.plugin_key = plugin_key

        if defaults is not None:
            self.defaults = {"plugins": {}}
            self.defaults["plugins"][plugin_key] = defaults
            self.defaults["plugins"][plugin_key]["_config_version"] = None
        else:
            self.defaults = None

        if get_preprocessors is None:
            get_preprocessors = {}
        self.get_preprocessors = {"plugins": {}}
        self.get_preprocessors["plugins"][plugin_key] = get_preprocessors

        if set_preprocessors is None:
            set_preprocessors = {}
        self.set_preprocessors = {"plugins": {}}
        self.set_preprocessors["plugins"][plugin_key] = set_preprocessors

        def prefix_path_in_args(args, index=0):
            result = []
            if index == 0:
                result.append(self._prefix_path(args[0]))
                result.extend(args[1:])
            else:
                args_before = args[: index - 1]
                args_after = args[index + 1 :]
                result.extend(args_before)
                result.append(self._prefix_path(args[index]))
                result.extend(args_after)
            return result

        def add_getter_kwargs(kwargs):
            if "defaults" not in kwargs and self.defaults is not None:
                kwargs.update(defaults=self.defaults)
            if "preprocessors" not in kwargs:
                kwargs.update(preprocessors=self.get_preprocessors)
            return kwargs

        def add_setter_kwargs(kwargs):
            if "defaults" not in kwargs and self.defaults is not None:
                kwargs.update(defaults=self.defaults)
            if "preprocessors" not in kwargs:
                kwargs.update(preprocessors=self.set_preprocessors)
            return kwargs

        def wrap_overlay(args):
            result = list(args)
            overlay = result[0]
            result[0] = {"plugins": {plugin_key: overlay}}
            return result

        self.access_methods = {
            "has": ("has", prefix_path_in_args, add_getter_kwargs),
            "get": ("get", prefix_path_in_args, add_getter_kwargs),
            "get_int": ("getInt", prefix_path_in_args, add_getter_kwargs),
            "get_float": ("getFloat", prefix_path_in_args, add_getter_kwargs),
            "get_boolean": ("getBoolean", prefix_path_in_args, add_getter_kwargs),
            "set": ("set", prefix_path_in_args, add_setter_kwargs),
            "set_int": ("setInt", prefix_path_in_args, add_setter_kwargs),
            "set_float": ("setFloat", prefix_path_in_args, add_setter_kwargs),
            "set_boolean": ("setBoolean", prefix_path_in_args, add_setter_kwargs),
            "remove": ("remove", prefix_path_in_args, lambda x: x),
            "add_overlay": ("add_overlay", wrap_overlay, lambda x: x),
            "remove_overlay": ("remove_overlay", lambda x: x, lambda x: x),
        }
        self.deprecated_access_methods = {
            "getInt": "get_int",
            "getFloat": "get_float",
            "getBoolean": "get_boolean",
            "setInt": "set_int",
            "setFloat": "set_float",
            "setBoolean": "set_boolean",
        }

    def _prefix_path(self, path=None):
        if path is None:
            path = list()
        return ["plugins", self.plugin_key] + path

    def global_has(self, path, **kwargs):
        return self.settings.has(path, **kwargs)

    def global_remove(self, path, **kwargs):
        return self.settings.remove(path, **kwargs)

    def global_get(self, path, **kwargs):
        """
        Getter for retrieving settings not managed by the plugin itself from the core settings structure. Use this
        to access global settings outside of your plugin.

        Directly forwards to :func:`octoprint.settings.Settings.get`.
        """
        return self.settings.get(path, **kwargs)

    def global_get_int(self, path, **kwargs):
        """
        Like :func:`global_get` but directly forwards to :func:`octoprint.settings.Settings.getInt`.
        """
        return self.settings.getInt(path, **kwargs)

    def global_get_float(self, path, **kwargs):
        """
        Like :func:`global_get` but directly forwards to :func:`octoprint.settings.Settings.getFloat`.
        """
        return self.settings.getFloat(path, **kwargs)

    def global_get_boolean(self, path, **kwargs):
        """
        Like :func:`global_get` but directly orwards to :func:`octoprint.settings.Settings.getBoolean`.
        """
        return self.settings.getBoolean(path, **kwargs)

    def global_set(self, path, value, **kwargs):
        """
        Setter for modifying settings not managed by the plugin itself on the core settings structure. Use this
        to modify global settings outside of your plugin.

        Directly forwards to :func:`octoprint.settings.Settings.set`.
        """
        self.settings.set(path, value, **kwargs)

    def global_set_int(self, path, value, **kwargs):
        """
        Like :func:`global_set` but directly forwards to :func:`octoprint.settings.Settings.setInt`.
        """
        self.settings.setInt(path, value, **kwargs)

    def global_set_float(self, path, value, **kwargs):
        """
        Like :func:`global_set` but directly forwards to :func:`octoprint.settings.Settings.setFloat`.
        """
        self.settings.setFloat(path, value, **kwargs)

    def global_set_boolean(self, path, value, **kwargs):
        """
        Like :func:`global_set` but directly forwards to :func:`octoprint.settings.Settings.setBoolean`.
        """
        self.settings.setBoolean(path, value, **kwargs)

    def global_get_basefolder(self, folder_type, **kwargs):
        """
        Retrieves a globally defined basefolder of the given ``folder_type``. Directly forwards to
        :func:`octoprint.settings.Settings.getBaseFolder`.
        """
        return self.settings.getBaseFolder(folder_type, **kwargs)

    def get_plugin_logfile_path(self, postfix=None):
        """
        Retrieves the path to a logfile specifically for the plugin. If ``postfix`` is not supplied, the logfile
        will be named ``plugin_<plugin identifier>.log`` and located within the configured ``logs`` folder. If a
        postfix is supplied, the name will be ``plugin_<plugin identifier>_<postfix>.log`` at the same location.

        Plugins may use this for specific logging tasks. For example, a :class:`~octoprint.plugin.SlicingPlugin` might
        want to create a log file for logging the output of the slicing engine itself if some debug flag is set.

        Arguments:
            postfix (str): Postfix of the logfile for which to create the path. If set, the file name of the log file
                will be ``plugin_<plugin identifier>_<postfix>.log``, if not it will be
                ``plugin_<plugin identifier>.log``.

        Returns:
            str: Absolute path to the log file, directly usable by the plugin.
        """
        filename = "plugin_" + self.plugin_key
        if postfix is not None:
            filename += "_" + postfix
        filename += ".log"
        return os.path.join(self.settings.getBaseFolder("logs"), filename)

    @deprecated(
        "PluginSettings.get_plugin_data_folder has been replaced by OctoPrintPlugin.get_plugin_data_folder",
        includedoc="Replaced by :func:`~octoprint.plugin.types.OctoPrintPlugin.get_plugin_data_folder`",
        since="1.2.0",
    )
    def get_plugin_data_folder(self):
        path = os.path.join(self.settings.getBaseFolder("data"), self.plugin_key)
        if not os.path.isdir(path):
            os.makedirs(path)
        return path

    def get_all_data(self, **kwargs):
        merged = kwargs.get("merged", True)
        asdict = kwargs.get("asdict", True)
        defaults = kwargs.get("defaults", self.defaults)
        preprocessors = kwargs.get("preprocessors", self.get_preprocessors)

        kwargs.update(
            {
                "merged": merged,
                "asdict": asdict,
                "defaults": defaults,
                "preprocessors": preprocessors,
            }
        )

        return self.settings.get(self._prefix_path(), **kwargs)

    def clean_all_data(self):
        self.settings.remove(self._prefix_path())

    def __getattr__(self, item):
        all_access_methods = list(self.access_methods.keys()) + list(
            self.deprecated_access_methods.keys()
        )
        if item in all_access_methods:
            decorator = None
            if item in self.deprecated_access_methods:
                new = self.deprecated_access_methods[item]
                decorator = deprecated(
                    "{old} has been renamed to {new}".format(old=item, new=new),
                    stacklevel=2,
                )
                item = new

            settings_name, args_mapper, kwargs_mapper = self.access_methods[item]
            if hasattr(self.settings, settings_name) and callable(
                getattr(self.settings, settings_name)
            ):
                orig_func = getattr(self.settings, settings_name)
                if decorator is not None:
                    orig_func = decorator(orig_func)

                def _func(*args, **kwargs):
                    return orig_func(*args_mapper(args), **kwargs_mapper(kwargs))

                _func.__name__ = to_native_str(item)
                _func.__doc__ = orig_func.__doc__ if "__doc__" in dir(orig_func) else None

                return _func

        return getattr(self.settings, item)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


def can_perform_update(target, check, online=True):
    return (
        "python_updater" in check
        and check["python_updater"] is not None
        and hasattr(check["python_updater"], "perform_update")
        and (online or check.get("offline", False))
    )


def perform_update(target, check, target_version, log_cb=None, online=True):
    from ..exceptions import CannotUpdateOffline

    if not online and not check("offline", False):
        raise CannotUpdateOffline()

    try:
        return check["python_updater"].perform_update(
            target, check, target_version, log_cb=log_cb, online=online
        )
    except Exception:
        import inspect

        args, _, _, _ = inspect.getargspec(check["python_updater"].perform_update)
        if "online" not in args:
            # old python_updater footprint, simply leave out the online parameter
            return check["python_updater"].perform_update(
                target, check, target_version, log_cb=log_cb
            )

        # some other error, raise again
        raise
